# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import Dict, Optional, Tuple, Union
from rdflib import RDF, RDFS, Graph
from rdflib.exceptions import UniquenessError
from rdflib.term import BNode, Node, URIRef

from spdx_tools.spdx.model import Snippet
from spdx_tools.spdx.parser.error import SPDXParsingError
from spdx_tools.spdx.parser.logger import Logger
from spdx_tools.spdx.parser.parsing_functions import (
    construct_or_raise_parsing_error,
    raise_parsing_error_if_logger_has_messages,
)
from spdx_tools.spdx.parser.rdf.graph_parsing_functions import (
    apply_parsing_method_or_log_error,
    get_correctly_typed_triples,
    get_correctly_typed_value,
    get_value_from_graph,
    parse_literal,
    parse_literal_or_no_assertion_or_none,
    parse_spdx_id,
)
from spdx_tools.spdx.parser.rdf.license_expression_parser import parse_license_expression
from spdx_tools.spdx.rdfschema.namespace import POINTER_NAMESPACE, SPDX_NAMESPACE


def parse_snippet(snippet_node: Union[URIRef, BNode], graph: Graph, doc_namespace: str) -> Snippet:
    logger = Logger()
    spdx_id = parse_spdx_id(snippet_node, doc_namespace, graph)
    file_spdx_id_uri = get_value_from_graph(
        logger, graph, subject=snippet_node, predicate=SPDX_NAMESPACE.snippetFromFile
    )
    file_spdx_id = parse_spdx_id(file_spdx_id_uri, doc_namespace, graph)
    byte_range = None
    line_range = None
    for _, _, start_end_pointer in graph.triples((snippet_node, SPDX_NAMESPACE.range, None)):
        parsed_range = apply_parsing_method_or_log_error(
            logger, start_end_pointer, parsing_method=lambda x: parse_ranges(x, graph)
        )
        byte_range, line_range = set_range_or_log_error(byte_range, line_range, logger, parsed_range)

    license_concluded = parse_literal_or_no_assertion_or_none(
        logger,
        graph,
        snippet_node,
        SPDX_NAMESPACE.licenseConcluded,
        parsing_method=lambda x: parse_license_expression(x, graph, doc_namespace, logger),
    )
    license_info_in_snippet = []
    for _, _, license_info_in_snippet_node in graph.triples((snippet_node, SPDX_NAMESPACE.licenseInfoInSnippet, None)):
        license_info_in_snippet.append(
            get_correctly_typed_value(
                logger,
                license_info_in_snippet_node,
                lambda x: parse_license_expression(x, graph, doc_namespace, logger),
            )
        )
    license_comment = parse_literal(logger, graph, snippet_node, SPDX_NAMESPACE.licenseComments)
    copyright_text = parse_literal_or_no_assertion_or_none(logger, graph, snippet_node, SPDX_NAMESPACE.copyrightText)
    comment = parse_literal(logger, graph, snippet_node, RDFS.comment)
    name = parse_literal(logger, graph, snippet_node, SPDX_NAMESPACE.name)
    attribution_texts = []
    for _, _, attribution_text_literal in get_correctly_typed_triples(
        logger, graph, snippet_node, SPDX_NAMESPACE.attributionText, None
    ):
        attribution_texts.append(attribution_text_literal.toPython())

    raise_parsing_error_if_logger_has_messages(logger, "Snippet")
    snippet = construct_or_raise_parsing_error(
        Snippet,
        dict(
            spdx_id=spdx_id,
            file_spdx_id=file_spdx_id,
            byte_range=byte_range,
            line_range=line_range,
            license_concluded=license_concluded,
            license_info_in_snippet=license_info_in_snippet,
            license_comment=license_comment,
            copyright_text=copyright_text,
            comment=comment,
            name=name,
            attribution_texts=attribution_texts,
        ),
    )
    return snippet


def set_range_or_log_error(
    byte_range: Optional[Tuple[int, int]],
    line_range: Optional[Tuple[int, int]],
    logger: Logger,
    parsed_range: Dict[str, Tuple[int, int]],
) -> Tuple[Optional[Tuple[int, int]], Optional[Tuple[int, int]]]:
    if not parsed_range:
        return byte_range, line_range
    if "ByteOffsetPointer" in parsed_range.keys() and not byte_range:
        byte_range = parsed_range["ByteOffsetPointer"]
    elif "ByteOffsetPointer" in parsed_range.keys() and byte_range:
        logger.append("Multiple ByteOffsetPointer found.")
    elif "LineCharPointer" in parsed_range.keys() and not line_range:
        line_range = parsed_range["LineCharPointer"]
    elif "LineCharPointer" in parsed_range.keys() and line_range:
        logger.append("Multiple LineCharPointer found.")
    return byte_range, line_range


def parse_ranges(start_end_pointer: BNode, graph: Graph) -> Dict[str, Tuple[int, int]]:
    range_values = dict()
    start_pointer_type, start_pointer_node = get_pointer_type(graph, POINTER_NAMESPACE.startPointer, start_end_pointer)
    end_pointer_type, end_pointer_node = get_pointer_type(graph, POINTER_NAMESPACE.endPointer, start_end_pointer)

    if start_pointer_type != end_pointer_type:
        raise SPDXParsingError(["Types of startPointer and endPointer don't match"])

    range_values["startPointer"] = parse_range_value(graph, start_pointer_node, POINTER_MATCHING[start_pointer_type])
    range_values["endPointer"] = parse_range_value(graph, end_pointer_node, POINTER_MATCHING[end_pointer_type])

    return {str(start_pointer_type.fragment): (range_values["startPointer"], range_values["endPointer"])}


def get_pointer_type(graph: Graph, pointer: URIRef, start_end_pointer: BNode) -> Tuple[URIRef, Node]:
    try:
        pointer_node = graph.value(start_end_pointer, pointer, any=False)
    except UniquenessError:
        raise SPDXParsingError([f"Multiple values for {pointer.fragment}"])
    if not pointer_node:
        raise SPDXParsingError([f"Couldn't find pointer of type {pointer.fragment}."])
    pointer_type = get_value_from_graph(Logger(), graph, pointer_node, RDF.type)
    return pointer_type, pointer_node


POINTER_MATCHING = {
    POINTER_NAMESPACE.ByteOffsetPointer: POINTER_NAMESPACE.offset,
    POINTER_NAMESPACE.LineCharPointer: POINTER_NAMESPACE.lineNumber,
}


def parse_range_value(graph: Graph, pointer_node: Node, predicate: URIRef) -> Optional[int]:
    try:
        value = get_value_from_graph(Logger(), graph, pointer_node, predicate, _any=False)
    except UniquenessError:
        raise SPDXParsingError([f"Multiple values for {predicate.fragment} found."])
    if value:
        value = int(value.toPython())
    return value



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import List, Optional, Union

from spdx_tools.spdx3.bump_from_spdx2.actor import bump_actor
from spdx_tools.spdx3.bump_from_spdx2.bump_utils import handle_no_assertion_or_none
from spdx_tools.spdx3.bump_from_spdx2.checksum import bump_checksum
from spdx_tools.spdx3.bump_from_spdx2.message import print_missing_conversion
from spdx_tools.spdx3.model import (
    ExternalIdentifier,
    ExternalIdentifierType,
    ExternalMap,
    ExternalReference,
    ExternalReferenceType,
)
from spdx_tools.spdx3.model.software import Package, SoftwarePurpose
from spdx_tools.spdx3.payload import Payload
from spdx_tools.spdx.model import Actor as Spdx2_Actor
from spdx_tools.spdx.model import ExternalDocumentRef, SpdxNoAssertion
from spdx_tools.spdx.model.package import ExternalPackageRef
from spdx_tools.spdx.model.package import Package as Spdx2_Package
from spdx_tools.spdx.spdx_element_utils import get_full_element_spdx_id


def bump_package(
    spdx2_package: Spdx2_Package,
    payload: Payload,
    document_namespace: str,
    external_document_refs: List[ExternalDocumentRef],
    imports: List[ExternalMap],
):
    spdx_id = get_full_element_spdx_id(spdx2_package, document_namespace, external_document_refs)
    if ":" in spdx2_package.spdx_id:
        imports.append(
            ExternalMap(
                external_id=spdx2_package.spdx_id,
                defining_document=f"{spdx2_package.spdx_id.split(':')[0]}:SPDXRef-DOCUMENT",
            )
        )

    download_location = handle_no_assertion_or_none(spdx2_package.download_location, "package.download_location")
    print_missing_conversion("package2.file_name", 0, "https://github.com/spdx/spdx-3-model/issues/83")
    if isinstance(spdx2_package.supplier, Spdx2_Actor):
        supplied_by_spdx_id = [bump_actor(spdx2_package.supplier, payload, document_namespace)]
    else:
        supplied_by_spdx_id = None
    if isinstance(spdx2_package.originator, Spdx2_Actor):
        originated_by_spdx_id = [bump_actor(spdx2_package.originator, payload, document_namespace)]
    else:
        originated_by_spdx_id = None
    print_missing_conversion("package2.files_analyzed", 0, "https://github.com/spdx/spdx-3-model/issues/84")
    print_missing_conversion(
        "package2.verification_code", 1, "of IntegrityMethod, https://github.com/spdx/spdx-3-model/issues/85"
    )
    integrity_methods = [bump_checksum(checksum) for checksum in spdx2_package.checksums]
    copyright_text = None
    if isinstance(spdx2_package.copyright_text, str):
        copyright_text = spdx2_package.copyright_text
    elif isinstance(spdx2_package.copyright_text, SpdxNoAssertion):
        print_missing_conversion("package2.copyright_text", 0)
    print_missing_conversion(
        "package2.license_info_from_files, package2.license_comment",
        0,
        "and missing definition of license profile",
    )

    external_reference = []
    external_identifier = []
    purl_refs = [
        external_ref for external_ref in spdx2_package.external_references if external_ref.reference_type == "purl"
    ]
    exactly_one_purl_without_comment = len(purl_refs) == 1 and purl_refs[0].comment is None
    package_url = None
    if exactly_one_purl_without_comment:
        package_url = purl_refs[0].locator
    for spdx2_external_ref in spdx2_package.external_references:
        if exactly_one_purl_without_comment and spdx2_external_ref.reference_type == "purl":
            continue
        id_or_ref = bump_external_package_ref(spdx2_external_ref)
        if isinstance(id_or_ref, ExternalReference):
            external_reference.append(id_or_ref)
        elif isinstance(id_or_ref, ExternalIdentifier):
            external_identifier.append(id_or_ref)

    package_purpose = (
        SoftwarePurpose[spdx2_package.primary_package_purpose.name] if spdx2_package.primary_package_purpose else None
    )

    payload.add_element(
        Package(
            spdx_id,
            spdx2_package.name,
            summary=spdx2_package.summary,
            description=spdx2_package.description,
            comment=spdx2_package.comment,
            verified_using=integrity_methods,
            external_reference=external_reference,
            external_identifier=external_identifier,
            originated_by=originated_by_spdx_id,
            supplied_by=supplied_by_spdx_id,
            built_time=spdx2_package.built_date,
            release_time=spdx2_package.release_date,
            valid_until_time=spdx2_package.valid_until_date,
            primary_purpose=package_purpose,
            package_version=spdx2_package.version,
            download_location=download_location,
            package_url=package_url,
            homepage=spdx2_package.homepage,
            source_info=spdx2_package.source_info,
            copyright_text=copyright_text,
            attribution_text=", ".join(spdx2_package.attribution_texts),
        )
    )


external_ref_type_map = {
    "cpe22Type": ExternalIdentifierType.CPE22,
    "cpe23Type": ExternalIdentifierType.CPE23,
    "advisory": ExternalReferenceType.SECURITY_ADVISORY,
    "fix": ExternalReferenceType.SECURITY_FIX,
    "url": None,
    "swid": ExternalIdentifierType.SWID,
    "maven-central": None,
    "npm": None,
    "nuget": None,
    "bower": None,
    "purl": ExternalIdentifierType.PURL,
    "swh": ExternalIdentifierType.SWHID,
    "gitoid": ExternalIdentifierType.GITOID,
}


def bump_external_package_ref(
    spdx2_external_ref: ExternalPackageRef,
) -> Optional[Union[ExternalReference, ExternalIdentifier]]:
    reference_type = spdx2_external_ref.reference_type
    locator = spdx2_external_ref.locator
    comment = spdx2_external_ref.comment

    if reference_type not in external_ref_type_map:
        print_missing_conversion(
            reference_type,
            0,
            f"Conversion of ExternalPackageRef of type {reference_type} is currently not supported."
            f"https://github.com/spdx/spdx-3-model/issues/81",
        )
        return None

    id_or_ref_type = external_ref_type_map[reference_type]

    if isinstance(id_or_ref_type, ExternalReferenceType):
        return ExternalReference(id_or_ref_type, [locator], None, comment)
    elif isinstance(id_or_ref_type, ExternalIdentifierType):
        return ExternalIdentifier(id_or_ref_type, locator, comment)



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from spdx_tools.spdx3.model.profile_identifier import ProfileIdentifier
from spdx_tools.spdx3.model.creation_info import CreationInfo
from spdx_tools.spdx3.model.integrity_method import IntegrityMethod
from spdx_tools.spdx3.model.hash import Hash, HashAlgorithm
from spdx_tools.spdx3.model.external_reference import ExternalReference, ExternalReferenceType
from spdx_tools.spdx3.model.external_identifier import ExternalIdentifier, ExternalIdentifierType
from spdx_tools.spdx3.model.external_map import ExternalMap
from spdx_tools.spdx3.model.namespace_map import NamespaceMap
from spdx_tools.spdx3.model.element import Element
from spdx_tools.spdx3.model.agent import Agent
from spdx_tools.spdx3.model.person import Person
from spdx_tools.spdx3.model.organization import Organization
from spdx_tools.spdx3.model.software_agent import SoftwareAgent
from spdx_tools.spdx3.model.tool import Tool
from spdx_tools.spdx3.model.spdx_collection import ElementCollection
from spdx_tools.spdx3.model.bundle import Bundle
from spdx_tools.spdx3.model.bom import Bom
from spdx_tools.spdx3.model.spdx_document import SpdxDocument
from spdx_tools.spdx3.model.annotation import Annotation, AnnotationType
from spdx_tools.spdx3.model.relationship import Relationship, RelationshipType, RelationshipCompleteness
from spdx_tools.spdx3.model.lifecycle_scoped_relationship import LifecycleScopedRelationship, LifecycleScopeType
from spdx_tools.spdx3.model.artifact import Artifact



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from unittest.mock import MagicMock, call, mock_open, patch

from spdx_tools.spdx.writer.tagvalue.annotation_writer import write_annotation
from tests.spdx.fixtures import annotation_fixture


def test_annotation_writer():
    annotation = annotation_fixture()

    mock: MagicMock = mock_open()
    with patch(f"{__name__}.open", mock, create=True):
        with open("foo", "w") as file:
            write_annotation(annotation, file)

    mock.assert_called_once_with("foo", "w")
    handle = mock()
    handle.write.assert_has_calls(
        [
            call(f"Annotator: Person: {annotation.annotator.name} ({annotation.annotator.email})\n"),
            call("AnnotationDate: 2022-12-24T00:00:00Z\n"),
            call(f"AnnotationType: {annotation.annotation_type.name}\n"),
            call(f"SPDXREF: {annotation.spdx_id}\n"),
            call(f"AnnotationComment: {annotation.annotation_comment}\n"),
        ]
    )



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

from typing import List

import pytest

from spdx_tools.spdx.model import Checksum, ChecksumAlgorithm
from spdx_tools.spdx.validation.checksum_validator import validate_checksum
from spdx_tools.spdx.validation.validation_message import SpdxElementType, ValidationContext, ValidationMessage
from tests.spdx.fixtures import checksum_fixture


@pytest.mark.parametrize(
    "checksum",
    [
        checksum_fixture(),
        Checksum(ChecksumAlgorithm.SHA1, "71c4025dd9897b364f3ebbb42c484ff43d00791c"),
        Checksum(ChecksumAlgorithm.SHA224, "9c9f4e27d957a123cc32d86afe33ae53b1184192cccb23b0f257f588"),
        Checksum(ChecksumAlgorithm.SHA256, "fbea580d286bbbbb41314430d58ba887716a74d7134119c5307cdc9f0c7a4299"),
        Checksum(
            ChecksumAlgorithm.SHA384,
            "73b4ad9a34e5f76cb2525ea6bb8b1dcf9ba79426b3295bd18bc6d148cba4fcc2ca3cf2630fd481b47caaac9127103933",
        ),
        Checksum(
            ChecksumAlgorithm.SHA512,
            "c2aa8a5d297f5e888ce9a30d3745ccc5a628533449a9f98524de3d23695a268f394a67faf8ef370727c2946f1dbbec34aeb7ac10f"
            "15af43e7cb5547f1a464053",
        ),
        Checksum(ChecksumAlgorithm.SHA3_256, "1e772489c042f49aeaae32b00fc5ef170a25afa741cffaafadde597d4d1727ce"),
        Checksum(
            ChecksumAlgorithm.SHA3_384,
            "dd9e30747551865b483bd76bd967384dce0e5670d1b1c3f701cffac7f49b1c46791253493835136b3aa5f679e364c166",
        ),
        Checksum(
            ChecksumAlgorithm.SHA3_512,
            "906bca5580be8c95ae44f775363fb69968ad568898dfb03e0ff96cd9445a0b75f817b68e5c1e80ad624031f851cfddd3a101e1d11"
            "1310266a5d46e2bc1ffbb36",
        ),
        Checksum(ChecksumAlgorithm.BLAKE2B_256, "a0eb3ddfa5807780a562b9c313b2537f1e8dc621e9a524f8c1ffcf07a79e35c7"),
        Checksum(
            ChecksumAlgorithm.BLAKE2B_384,
            "902511afc8939c0193d87857f45a19eddfd7e0413b0f8701a3baaf1b025f882b45a8fbf623fa0ad79b64850ac7a4d0b2",
        ),
        Checksum(
            ChecksumAlgorithm.BLAKE2B_512,
            "72c23b0160e1af3cb159f0cc96210c5e9aecc5a65d4618566776fa6117bf84929dcef56c7f8b087691c23000c945470842d90b5e8"
            "c4af74dce531ca8ebd8824c",
        ),
        Checksum(
            ChecksumAlgorithm.BLAKE3,
            "a872cac2efd29ed2ad8b5faa79b63f983341bea41183582b8863d952f6ac3e1cdfe0189967a13006857d3b9985174bf67239874d"
            "cec4cbbc9839496179feafeda872cac2efd29ed2ad8b5faa79b63f983341bea41183582b8863d952f6ac3e1cdfe0189967a13006"
            "857d3b9985174bf67239874dcec4cbbc9839496179feafed",
        ),
        Checksum(ChecksumAlgorithm.MD2, "af1eec2a1b18886c3f3cc244349d91d8"),
        Checksum(ChecksumAlgorithm.MD4, "d4c41ce30a517d6ce9d79c8c17bb4b66"),
        Checksum(ChecksumAlgorithm.MD5, "0d7f61beb7018b3924c6b8f96549fa39"),
        Checksum(
            ChecksumAlgorithm.MD6,
            "af1eec2a1b18886c3f3cc244349d91d8d4c41ce30a517d6ce9d79c8c17bb4b660d7f61beb7018b3924c6b8f96549fa39",
        ),
        Checksum(ChecksumAlgorithm.ADLER32, "02ec0130"),
    ],
)
def test_valid_checksum(checksum):
    validation_messages: List[ValidationMessage] = validate_checksum(checksum, "parent_id", "SPDX-2.3")

    assert validation_messages == []


@pytest.mark.parametrize(
    "checksum, expected_message",
    [
        (
            Checksum(ChecksumAlgorithm.SHA1, "af1eec2a1b18886c3f3cc244349d91d8"),
            "value of ChecksumAlgorithm.SHA1 must consist of 40 lowercase hexadecimal digits, but is: "
            "af1eec2a1b18886c3f3cc244349d91d8 (length: 32 digits)",
        ),
        (
            Checksum(ChecksumAlgorithm.SHA224, "af1eec2a1b18886c3f3cc244349d91d8"),
            "value of ChecksumAlgorithm.SHA224 must consist of 56 lowercase hexadecimal digits, but is: "
            "af1eec2a1b18886c3f3cc244349d91d8 (length: 32 digits)",
        ),
        (
            Checksum(ChecksumAlgorithm.SHA256, "af1eec2a1b18886c3f3cc244349d91d8"),
            "value of ChecksumAlgorithm.SHA256 must consist of 64 lowercase hexadecimal digits, but is: "
            "af1eec2a1b18886c3f3cc244349d91d8 (length: 32 digits)",
        ),
        (
            Checksum(ChecksumAlgorithm.SHA384, "af1eec2a1b18886c3f3cc244349d91d8"),
            "value of ChecksumAlgorithm.SHA384 must consist of 96 lowercase hexadecimal digits, but is: "
            "af1eec2a1b18886c3f3cc244349d91d8 (length: 32 digits)",
        ),
        (
            Checksum(ChecksumAlgorithm.SHA512, "af1eec2a1b18886c3f3cc244349d91d8"),
            "value of ChecksumAlgorithm.SHA512 must consist of 128 lowercase hexadecimal digits, but is: "
            "af1eec2a1b18886c3f3cc244349d91d8 (length: 32 digits)",
        ),
        (
            Checksum(ChecksumAlgorithm.SHA3_256, "af1eec2a1b18886c3f3cc244349d91d8"),
            "value of ChecksumAlgorithm.SHA3_256 must consist of 64 lowercase hexadecimal digits, but is: "
            "af1eec2a1b18886c3f3cc244349d91d8 (length: 32 digits)",
        ),
        (
            Checksum(ChecksumAlgorithm.SHA3_384, "af1eec2a1b18886c3f3cc244349d91d8"),
            "value of ChecksumAlgorithm.SHA3_384 must consist of 96 lowercase hexadecimal digits, but is: "
            "af1eec2a1b18886c3f3cc244349d91d8 (length: 32 digits)",
        ),
        (
            Checksum(ChecksumAlgorithm.SHA3_512, "af1eec2a1b18886c3f3cc244349d91d8"),
            "value of ChecksumAlgorithm.SHA3_512 must consist of 128 lowercase hexadecimal digits, but is: "
            "af1eec2a1b18886c3f3cc244349d91d8 (length: 32 digits)",
        ),
        (
            Checksum(ChecksumAlgorithm.BLAKE2B_256, "af1eec2a1b18886c3f3cc244349d91d8"),
            "value of ChecksumAlgorithm.BLAKE2B_256 must consist of 64 lowercase hexadecimal digits, but is: "
            "af1eec2a1b18886c3f3cc244349d91d8 (length: 32 digits)",
        ),
        (
            Checksum(ChecksumAlgorithm.BLAKE2B_384, "af1eec2a1b18886c3f3cc244349d91d8"),
            "value of ChecksumAlgorithm.BLAKE2B_384 must consist of 96 lowercase hexadecimal digits, but is: "
            "af1eec2a1b18886c3f3cc244349d91d8 (length: 32 digits)",
        ),
        (
            Checksum(ChecksumAlgorithm.BLAKE2B_512, "af1eec2a1b18886c3f3cc244349d91d8"),
            "value of ChecksumAlgorithm.BLAKE2B_512 must consist of 128 lowercase hexadecimal digits, but is: "
            "af1eec2a1b18886c3f3cc244349d91d8 (length: 32 digits)",
        ),
        (
            Checksum(ChecksumAlgorithm.BLAKE3, "af1eec2a1b18886c3f3cc244349d91d8"),
            "value of ChecksumAlgorithm.BLAKE3 must consist of at least 256 lowercase hexadecimal digits, but is: af1e"
            "ec2a1b18886c3f3cc244349d91d8 (length: 32 digits)",
        ),
        (
            Checksum(ChecksumAlgorithm.MD2, "71c4025dd9897b364f3ebbb42c484ff43d00791c"),
            "value of ChecksumAlgorithm.MD2 must consist of 32 lowercase hexadecimal digits, but is: 71c4025dd9897b364"
            "f3ebbb42c484ff43d00791c (length: 40 digits)",
        ),
        (
            Checksum(ChecksumAlgorithm.MD4, "71c4025dd9897b364f3ebbb42c484ff43d00791c"),
            "value of ChecksumAlgorithm.MD4 must consist of 32 lowercase hexadecimal digits, but is: 71c4025dd9897b364"
            "f3ebbb42c484ff43d00791c (length: 40 digits)",
        ),
        (
            Checksum(ChecksumAlgorithm.MD5, "71c4025dd9897b364f3ebbb42c484ff43d00791c"),
            "value of ChecksumAlgorithm.MD5 must consist of 32 lowercase hexadecimal digits, but is: 71c4025dd9897b364"
            "f3ebbb42c484ff43d00791c (length: 40 digits)",
        ),
        (
            Checksum(
                ChecksumAlgorithm.MD6,
                "a872cac2efd29ed2ad8b5faa79b63f983341bea41183582b8863d952f6ac3e1cdfe0189967a13006857d3b9985174bf672398"
                "74dcec4cbbc9839496179feafeda872cac2efd29ed2ad8b5faa79b63f983341bea41183582b8863d952f6ac3e1cdfe0189967"
                "a13006857d3b9985174bf67239874dcec4cbbc9839496179feafeda872cac2efd29ed2ad8b5faa79b63f983341bea41183582"
                "b8863d952f6ac3e1cdfe0189967a13006857d3b9985174bf67239874dcec4cbbc9839496179feafeda872cac2efd29ed2ad8b"
                "5faa79b63f983341bea41183582b8863d952f6ac3e1cdfe0189967a13006857d3b9985174bf67239874dcec4cbbc983949617"
                "9feafed5",
            ),
            "value of ChecksumAlgorithm.MD6 must consist of between 0 and 512 lowercase hexadecimal digits, but is: "
            "a872cac2efd29ed2ad8b5faa79b63f983341bea41183582b8863d952f6ac3e1cdfe0189967a13006857d3b9985174bf67239874dc"
            "ec4cbbc9839496179feafeda872cac2efd29ed2ad8b5faa79b63f983341bea41183582b8863d952f6ac3e1cdfe0189967a1300685"
            "7d3b9985174bf67239874dcec4cbbc9839496179feafeda872cac2efd29ed2ad8b5faa79b63f983341bea41183582b8863d952f6a"
            "c3e1cdfe0189967a13006857d3b9985174bf67239874dcec4cbbc9839496179feafeda872cac2efd29ed2ad8b5faa79b63f983341"
            "bea41183582b8863d952f6ac3e1cdfe0189967a13006857d3b9985174bf67239874dcec4cbbc9839496179feafed5 "
            "(length: 513 digits)",
        ),
        (
            Checksum(ChecksumAlgorithm.ADLER32, "af1eec2a1b18886c3f3cc244349d91d8"),
            "value of ChecksumAlgorithm.ADLER32 must consist of 8 lowercase hexadecimal digits, but is: "
            "af1eec2a1b18886c3f3cc244349d91d8 (length: 32 digits)",
        ),
        (
            Checksum(ChecksumAlgorithm.SHA1, "CE9F343C4BA371746FD7EAD9B59031AE34D8AFC4"),
            "value of ChecksumAlgorithm.SHA1 must consist of 40 lowercase hexadecimal digits, but is: "
            "CE9F343C4BA371746FD7EAD9B59031AE34D8AFC4 (length: 40 digits)",
        ),
    ],
)
def test_invalid_checksum(checksum, expected_message):
    parent_id = "parent_id"
    validation_messages: List[ValidationMessage] = validate_checksum(checksum, parent_id, "SPDX-2.3")

    expected = ValidationMessage(
        expected_message,
        ValidationContext(parent_id=parent_id, element_type=SpdxElementType.CHECKSUM, full_element=checksum),
    )

    assert validation_messages == [expected]


@pytest.mark.parametrize(
    "checksum",
    [
        Checksum(ChecksumAlgorithm.SHA3_256, "1e772489c042f49aeaae32b00fc5ef170a25afa741cffaafadde597d4d1727ce"),
        Checksum(
            ChecksumAlgorithm.SHA3_384,
            "dd9e30747551865b483bd76bd967384dce0e5670d1b1c3f701cffac7f49b1c46791253493835136b3aa5f679e364c166",
        ),
        Checksum(
            ChecksumAlgorithm.SHA3_512,
            "906bca5580be8c95ae44f775363fb69968ad568898dfb03e0ff96cd9445a0b75f817b68e5c1e80ad624031f851cfddd3a101e1d1"
            "11310266a5d46e2bc1ffbb36",
        ),
        Checksum(ChecksumAlgorithm.BLAKE2B_256, "a0eb3ddfa5807780a562b9c313b2537f1e8dc621e9a524f8c1ffcf07a79e35c7"),
        Checksum(
            ChecksumAlgorithm.BLAKE2B_384,
            "902511afc8939c0193d87857f45a19eddfd7e0413b0f8701a3baaf1b025f882b45a8fbf623fa0ad79b64850ac7a4d0b2",
        ),
        Checksum(
            ChecksumAlgorithm.BLAKE2B_512,
            "72c23b0160e1af3cb159f0cc96210c5e9aecc5a65d4618566776fa6117bf84929dcef56c7f8b087691c23000c945470842d90b5e"
            "8c4af74dce531ca8ebd8824c",
        ),
        Checksum(
            ChecksumAlgorithm.BLAKE3,
            "a872cac2efd29ed2ad8b5faa79b63f983341bea41183582b8863d952f6ac3e1cdfe0189967a13006857d3b9985174bf67239874"
            "dcec4cbbc9839496179feafeda872cac2efd29ed2ad8b5faa79b63f983341bea41183582b8863d952f6ac3e1cdfe0189967a13006"
            "857d3b9985174bf67239874dcec4cbbc9839496179feafed",
        ),
        Checksum(ChecksumAlgorithm.ADLER32, "02ec0130"),
    ],
)
def test_v2_3only_checksums(checksum):
    parent_id = "parent_id"
    validation_messages: List[ValidationMessage] = validate_checksum(checksum, parent_id, "SPDX-2.2")

    context = ValidationContext(parent_id=parent_id, element_type=SpdxElementType.CHECKSUM, full_element=checksum)
    expected = ValidationMessage(f"{checksum.algorithm.name} is not supported in SPDX-2.2", context)

    assert validation_messages == [expected]



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

import re

from beartype.typing import List

from spdx_tools.spdx.document_utils import get_contained_spdx_element_ids
from spdx_tools.spdx.model import Document, File


def is_valid_internal_spdx_id(spdx_id: str) -> bool:
    return bool(re.match(r"^SPDXRef-[\da-zA-Z.-]+$", spdx_id))


def is_valid_external_doc_ref_id(external_ref_id: str) -> bool:
    return bool(re.match(r"^DocumentRef-[\da-zA-Z.+-]+$", external_ref_id))


def is_spdx_id_present_in_files(spdx_id: str, files: List[File]) -> bool:
    return spdx_id in [file.spdx_id for file in files]


def is_spdx_id_present_in_document(spdx_id: str, document: Document) -> bool:
    all_spdx_ids_in_document: List[str] = get_list_of_all_spdx_ids(document)

    return spdx_id in all_spdx_ids_in_document


def get_list_of_all_spdx_ids(document: Document) -> List[str]:
    all_spdx_ids_in_document: List[str] = [document.creation_info.spdx_id]
    all_spdx_ids_in_document.extend(get_contained_spdx_element_ids(document))

    return all_spdx_ids_in_document


def is_external_doc_ref_present_in_document(external_ref_id: str, document: Document) -> bool:
    all_external_doc_ref_ids_in_document = [
        external_doc_ref.document_ref_id for external_doc_ref in document.creation_info.external_document_refs
    ]

    return external_ref_id in all_external_doc_ref_ids_in_document


def validate_spdx_id(
    spdx_id: str, document: Document, check_document: bool = False, check_files: bool = False
) -> List[str]:
    """Test that the given spdx_id (and a potential DocumentRef to an external document) is valid
    and, if it is a reference, actually exists in the document. Optionally checks files or the whole document
    for the existence of the spdx_id (i.e. if it is used as a reference). Returns a list of validation messages."""

    validation_messages: List[str] = []
    split_id: List[str] = spdx_id.split(":")

    # # # invalid case # # #
    if len(split_id) > 2:
        return [
            f"spdx_id must not contain more than one colon in order to separate the external document reference id "
            f"from the internal SPDX id, but is: {spdx_id}"
        ]

    # # # case with external document ref prefix # # #
    if len(split_id) == 2:
        if not is_valid_external_doc_ref_id(split_id[0]):
            validation_messages.append(
                f'the external document reference part of spdx_id must only contain letters, numbers, ".", "-" and '
                f'"+" and must begin with "DocumentRef-", but is: {split_id[0]}'
            )
        if not is_valid_internal_spdx_id(split_id[1]):
            validation_messages.append(
                f'the internal SPDX id part of spdx_id must only contain letters, numbers, "." and "-" and must begin '
                f'with "SPDXRef-", but is: {split_id[1]}'
            )
        if not is_external_doc_ref_present_in_document(split_id[0], document):
            validation_messages.append(
                f'did not find the external document reference "{split_id[0]}" in the SPDX document'
            )

        return validation_messages

    # # # "normal" case # # #
    if not is_valid_internal_spdx_id(spdx_id):
        validation_messages.append(
            f'spdx_id must only contain letters, numbers, "." and "-" and must begin with "SPDXRef-", but is: '
            f"{spdx_id}"
        )

    if check_document:
        if not is_spdx_id_present_in_document(spdx_id, document):
            validation_messages.append(f'did not find the referenced spdx_id "{spdx_id}" in the SPDX document')

    if check_files:
        if not is_spdx_id_present_in_files(spdx_id, document.files):
            validation_messages.append(
                f'did not find the referenced spdx_id "{spdx_id}" in the SPDX document\'s files'
            )

    return validation_messages



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import Dict, List, Optional, Union

from spdx_tools.spdx.model import ExtractedLicensingInfo, SpdxNoAssertion
from spdx_tools.spdx.parser.jsonlikedict.dict_parsing_functions import parse_field_or_no_assertion
from spdx_tools.spdx.parser.logger import Logger
from spdx_tools.spdx.parser.parsing_functions import construct_or_raise_parsing_error


class ExtractedLicensingInfoParser:
    logger: Logger

    def __init__(self):
        self.logger = Logger()

    @staticmethod
    def parse_extracted_licensing_info(extracted_licensing_info_dict: Dict) -> ExtractedLicensingInfo:
        license_id: Optional[str] = extracted_licensing_info_dict.get("licenseId")
        extracted_text: Optional[str] = extracted_licensing_info_dict.get("extractedText")
        license_name: Optional[Union[str, SpdxNoAssertion]] = parse_field_or_no_assertion(
            extracted_licensing_info_dict.get("name")
        )
        cross_references: List[str] = extracted_licensing_info_dict.get("seeAlsos", [])
        comment: Optional[str] = extracted_licensing_info_dict.get("comment")
        extracted_licensing_info = construct_or_raise_parsing_error(
            ExtractedLicensingInfo,
            dict(
                license_id=license_id,
                extracted_text=extracted_text,
                comment=comment,
                license_name=license_name,
                cross_references=cross_references,
            ),
        )
        return extracted_licensing_info



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import Any, Type

from spdx_tools.spdx.document_utils import get_contained_spdx_element_ids
from spdx_tools.spdx.jsonschema.annotation_converter import AnnotationConverter
from spdx_tools.spdx.jsonschema.converter import TypedConverter
from spdx_tools.spdx.jsonschema.creation_info_converter import CreationInfoConverter
from spdx_tools.spdx.jsonschema.document_properties import DocumentProperty
from spdx_tools.spdx.jsonschema.external_document_ref_converter import ExternalDocumentRefConverter
from spdx_tools.spdx.jsonschema.extracted_licensing_info_converter import ExtractedLicensingInfoConverter
from spdx_tools.spdx.jsonschema.file_converter import FileConverter
from spdx_tools.spdx.jsonschema.json_property import JsonProperty
from spdx_tools.spdx.jsonschema.package_converter import PackageConverter
from spdx_tools.spdx.jsonschema.relationship_converter import RelationshipConverter
from spdx_tools.spdx.jsonschema.snippet_converter import SnippetConverter
from spdx_tools.spdx.model import Document


class DocumentConverter(TypedConverter[Document]):
    creation_info_converter: CreationInfoConverter
    external_document_ref_converter: ExternalDocumentRefConverter
    package_converter: PackageConverter
    file_converter: FileConverter
    snippet_converter: SnippetConverter
    annotation_converter: AnnotationConverter
    relationship_converter: RelationshipConverter
    extracted_licensing_info_converter: ExtractedLicensingInfoConverter

    def __init__(self):
        self.external_document_ref_converter = ExternalDocumentRefConverter()
        self.creation_info_converter = CreationInfoConverter()
        self.package_converter = PackageConverter()
        self.file_converter = FileConverter()
        self.snippet_converter = SnippetConverter()
        self.annotation_converter = AnnotationConverter()
        self.relationship_converter = RelationshipConverter()
        self.extracted_licensing_info_converter = ExtractedLicensingInfoConverter()

    def get_json_type(self) -> Type[JsonProperty]:
        return DocumentProperty

    def get_data_model_type(self) -> Type[Document]:
        return Document

    def json_property_name(self, document_property: DocumentProperty) -> str:
        if document_property == DocumentProperty.SPDX_ID:
            return "SPDXID"
        return super().json_property_name(document_property)

    def _get_property_value(
        self, document: Document, document_property: DocumentProperty, _document: Document = None
    ) -> Any:
        if document_property == DocumentProperty.SPDX_ID:
            return document.creation_info.spdx_id
        elif document_property == DocumentProperty.ANNOTATIONS:
            # annotations referencing files, packages or snippets will be added to those elements directly
            element_ids = get_contained_spdx_element_ids(document)
            document_annotations = filter(
                lambda annotation: annotation.spdx_id not in element_ids, document.annotations
            )
            return [self.annotation_converter.convert(annotation) for annotation in document_annotations] or None
        elif document_property == DocumentProperty.COMMENT:
            return document.creation_info.document_comment
        elif document_property == DocumentProperty.CREATION_INFO:
            return self.creation_info_converter.convert(document.creation_info)
        elif document_property == DocumentProperty.DATA_LICENSE:
            return document.creation_info.data_license
        elif document_property == DocumentProperty.EXTERNAL_DOCUMENT_REFS:
            return [
                self.external_document_ref_converter.convert(external_document_ref)
                for external_document_ref in document.creation_info.external_document_refs
            ] or None
        elif document_property == DocumentProperty.HAS_EXTRACTED_LICENSING_INFOS:
            return [
                self.extracted_licensing_info_converter.convert(licensing_info)
                for licensing_info in document.extracted_licensing_info
            ] or None
        elif document_property == DocumentProperty.NAME:
            return document.creation_info.name
        elif document_property == DocumentProperty.SPDX_VERSION:
            return document.creation_info.spdx_version
        elif document_property == DocumentProperty.DOCUMENT_NAMESPACE:
            return document.creation_info.document_namespace
        elif document_property == DocumentProperty.PACKAGES:
            return [self.package_converter.convert(package, document) for package in document.packages] or None
        elif document_property == DocumentProperty.FILES:
            return [self.file_converter.convert(file, document) for file in document.files] or None
        elif document_property == DocumentProperty.SNIPPETS:
            return [self.snippet_converter.convert(snippet, document) for snippet in document.snippets] or None
        elif document_property == DocumentProperty.RELATIONSHIPS:
            return [
                self.relationship_converter.convert(relationship) for relationship in document.relationships
            ] or None



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import pytest

from spdx_tools.spdx.jsonschema.extracted_licensing_info_converter import ExtractedLicensingInfoConverter
from spdx_tools.spdx.jsonschema.extracted_licensing_info_properties import ExtractedLicensingInfoProperty
from spdx_tools.spdx.model import ExtractedLicensingInfo, SpdxNoAssertion
from spdx_tools.spdx.model.spdx_no_assertion import SPDX_NO_ASSERTION_STRING
from tests.spdx.fixtures import extracted_licensing_info_fixture


@pytest.fixture
def converter() -> ExtractedLicensingInfoConverter:
    return ExtractedLicensingInfoConverter()


@pytest.mark.parametrize(
    "extracted_licensing_info_property,expected",
    [
        (ExtractedLicensingInfoProperty.LICENSE_ID, "licenseId"),
        (ExtractedLicensingInfoProperty.EXTRACTED_TEXT, "extractedText"),
        (ExtractedLicensingInfoProperty.NAME, "name"),
        (ExtractedLicensingInfoProperty.COMMENT, "comment"),
        (ExtractedLicensingInfoProperty.SEE_ALSOS, "seeAlsos"),
    ],
)
def test_json_property_names(
    converter: ExtractedLicensingInfoConverter,
    extracted_licensing_info_property: ExtractedLicensingInfoProperty,
    expected: str,
):
    assert converter.json_property_name(extracted_licensing_info_property) == expected


def test_json_type(converter: ExtractedLicensingInfoConverter):
    assert converter.get_json_type() == ExtractedLicensingInfoProperty


def test_data_model_type(converter: ExtractedLicensingInfoConverter):
    assert converter.get_data_model_type() == ExtractedLicensingInfo


def test_successful_conversion(converter: ExtractedLicensingInfoConverter):
    extracted_licensing_info = ExtractedLicensingInfo(
        license_id="licenseId",
        extracted_text="Extracted text",
        license_name="license name",
        cross_references=["reference1", "reference2"],
        comment="comment",
    )

    converted_dict = converter.convert(extracted_licensing_info)

    assert converted_dict == {
        converter.json_property_name(ExtractedLicensingInfoProperty.LICENSE_ID): "licenseId",
        converter.json_property_name(ExtractedLicensingInfoProperty.EXTRACTED_TEXT): "Extracted text",
        converter.json_property_name(ExtractedLicensingInfoProperty.NAME): "license name",
        converter.json_property_name(ExtractedLicensingInfoProperty.SEE_ALSOS): ["reference1", "reference2"],
        converter.json_property_name(ExtractedLicensingInfoProperty.COMMENT): "comment",
    }


def test_null_values(converter: ExtractedLicensingInfoConverter):
    extracted_licensing_info = ExtractedLicensingInfo(cross_references=[])

    converted_dict = converter.convert(extracted_licensing_info)

    assert converter.json_property_name(ExtractedLicensingInfoProperty.LICENSE_ID) not in converted_dict
    assert converter.json_property_name(ExtractedLicensingInfoProperty.EXTRACTED_TEXT) not in converted_dict
    assert converter.json_property_name(ExtractedLicensingInfoProperty.NAME) not in converted_dict
    assert converter.json_property_name(ExtractedLicensingInfoProperty.SEE_ALSOS) not in converted_dict
    assert converter.json_property_name(ExtractedLicensingInfoProperty.COMMENT) not in converted_dict


def test_spdx_no_assertion(converter: ExtractedLicensingInfoConverter):
    extracted_licensing_info = extracted_licensing_info_fixture(license_name=SpdxNoAssertion())

    converted_dict = converter.convert(extracted_licensing_info)

    assert (
        converted_dict[converter.json_property_name(ExtractedLicensingInfoProperty.NAME)] == SPDX_NO_ASSERTION_STRING
    )



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from unittest import TestCase

import pytest
from license_expression import get_spdx_licensing

from spdx_tools.spdx.model import SpdxNoAssertion, SpdxNone
from spdx_tools.spdx.parser.error import SPDXParsingError
from spdx_tools.spdx.parser.jsonlikedict.license_expression_parser import LicenseExpressionParser


@pytest.mark.parametrize(
    "license_expression_str, expected_license",
    [
        ("First License", get_spdx_licensing().parse("First License")),
        ("Second License", get_spdx_licensing().parse("Second License")),
        ("NOASSERTION", SpdxNoAssertion()),
        ("NONE", SpdxNone()),
    ],
)
def test_parse_license_expression(license_expression_str, expected_license):
    license_expression_parser = LicenseExpressionParser()
    license_expression = license_expression_parser.parse_license_expression(license_expression_str)

    assert license_expression == expected_license


@pytest.mark.parametrize(
    "invalid_license_expression,expected_message",
    [
        (56, ["Error parsing LicenseExpression: expression must be a string and not: <class 'int'>: 56"]),
    ],
)
def test_parse_invalid_license_expression(invalid_license_expression, expected_message):
    license_expression_parser = LicenseExpressionParser()

    with pytest.raises(SPDXParsingError) as err:
        license_expression_parser.parse_license_expression(invalid_license_expression)

    TestCase().assertCountEqual(err.value.get_messages(), expected_message)



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from abc import abstractmethod
from dataclasses import field

from beartype.typing import List

from spdx_tools.common.typing.dataclass_with_properties import dataclass_with_properties
from spdx_tools.spdx3.model import Element, ExternalMap, NamespaceMap


@dataclass_with_properties
class ElementCollection(Element):
    # due to the inheritance we need to make all fields non-default in the __annotation__,
    # the __init__ method still raises an error if required fields are not set
    element: List[str] = field(default_factory=list)
    root_element: List[str] = field(default_factory=list)
    namespaces: List[NamespaceMap] = field(default_factory=list)
    imports: List[ExternalMap] = field(default_factory=list)

    @abstractmethod
    def __init__(self):
        pass



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from typing import Any, List, Type


def get_property_names(clazz: Type[Any]) -> List[str]:
    return [
        attribute
        for attribute in dir(clazz)
        if not attribute.startswith("_") and not callable(getattr(clazz, attribute))
    ]


class InvalidTypeClass:
    pass



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import pytest

from spdx_tools.spdx.jsonschema.relationship_converter import RelationshipConverter
from spdx_tools.spdx.jsonschema.relationship_properties import RelationshipProperty
from spdx_tools.spdx.model import Relationship, RelationshipType, SpdxNoAssertion, SpdxNone
from spdx_tools.spdx.model.spdx_no_assertion import SPDX_NO_ASSERTION_STRING
from spdx_tools.spdx.model.spdx_none import SPDX_NONE_STRING
from tests.spdx.fixtures import relationship_fixture


@pytest.fixture
def converter() -> RelationshipConverter:
    return RelationshipConverter()


@pytest.mark.parametrize(
    "relationship_property,expected",
    [
        (RelationshipProperty.SPDX_ELEMENT_ID, "spdxElementId"),
        (RelationshipProperty.COMMENT, "comment"),
        (RelationshipProperty.RELATED_SPDX_ELEMENT, "relatedSpdxElement"),
        (RelationshipProperty.RELATIONSHIP_TYPE, "relationshipType"),
    ],
)
def test_json_property_names(
    converter: RelationshipConverter, relationship_property: RelationshipProperty, expected: str
):
    assert converter.json_property_name(relationship_property) == expected


def test_json_type(converter: RelationshipConverter):
    assert converter.get_json_type() == RelationshipProperty


def test_data_model_type(converter: RelationshipConverter):
    assert converter.get_data_model_type() == Relationship


def test_successful_conversion(converter: RelationshipConverter):
    relationship = Relationship("spdxElementId", RelationshipType.COPY_OF, "relatedElementId", "comment")

    converted_dict = converter.convert(relationship)

    assert converted_dict == {
        converter.json_property_name(RelationshipProperty.SPDX_ELEMENT_ID): "spdxElementId",
        converter.json_property_name(RelationshipProperty.COMMENT): "comment",
        converter.json_property_name(RelationshipProperty.RELATED_SPDX_ELEMENT): "relatedElementId",
        converter.json_property_name(RelationshipProperty.RELATIONSHIP_TYPE): "COPY_OF",
    }


def test_spdx_no_assertion(converter: RelationshipConverter):
    relationship = relationship_fixture(related_spdx_element_id=SpdxNoAssertion())

    converted_dict = converter.convert(relationship)

    assert (
        converted_dict[converter.json_property_name(RelationshipProperty.RELATED_SPDX_ELEMENT)]
        == SPDX_NO_ASSERTION_STRING
    )


def test_spdx_none(converter: RelationshipConverter):
    relationship = relationship_fixture(related_spdx_element_id=SpdxNone())

    converted_dict = converter.convert(relationship)

    assert converted_dict[converter.json_property_name(RelationshipProperty.RELATED_SPDX_ELEMENT)] == SPDX_NONE_STRING



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import TextIO

from spdx_tools.spdx3.model import ExternalIdentifier
from spdx_tools.spdx3.writer.console.console import write_value


def write_external_identifier(external_identifier: ExternalIdentifier, text_output: TextIO):
    for property_name in ExternalIdentifier.__annotations__.keys():
        write_value(property_name, getattr(external_identifier, property_name), text_output)



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import pytest

from spdx_tools.spdx.writer.rdf.writer_utils import add_namespace_to_spdx_id


@pytest.mark.parametrize(
    "spdx_id,namespace,external_namespaces,expected",
    [
        ("SPDXRef-File", "docNamespace", {}, "docNamespace#SPDXRef-File"),
        (
            "externalDoc:SPDXRef-File",
            "docNamespace",
            {"externalDoc": "externalNamespace"},
            "externalNamespace#SPDXRef-File",
        ),
        ("externalDoc#A-Ref", "", {}, "externalDoc#A-Ref"),
        ("externalDoc:A-Ref", "", {}, "externalDoc:A-Ref"),
    ],
)
def test_add_namespace_to_spdx_id(spdx_id, namespace, expected, external_namespaces):
    extended_spdx_id = add_namespace_to_spdx_id(spdx_id, namespace, external_namespaces)

    assert extended_spdx_id == expected



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import os
from datetime import datetime

from rdflib import BNode, Graph, URIRef

from spdx_tools.spdx.model import Actor, ActorType, AnnotationType
from spdx_tools.spdx.parser.rdf.annotation_parser import parse_annotation
from spdx_tools.spdx.rdfschema.namespace import SPDX_NAMESPACE


def test_parse_annotation():
    graph = Graph().parse(os.path.join(os.path.dirname(__file__), "data/file_to_test_rdf_parser.rdf.xml"))
    doc_namespace = "https://some.namespace"
    file_node = URIRef(f"{doc_namespace}#SPDXRef-File")
    annotation_node = graph.value(subject=file_node, predicate=SPDX_NAMESPACE.annotation)
    assert isinstance(annotation_node, BNode)

    annotation = parse_annotation(annotation_node, graph, file_node, doc_namespace)

    assert annotation.spdx_id == "SPDXRef-File"
    assert annotation.annotation_type == AnnotationType.REVIEW
    assert annotation.annotator == Actor(ActorType.PERSON, "annotatorName", "some@mail.com")
    assert annotation.annotation_date == datetime(2022, 12, 1, 0, 0)
    assert annotation.annotation_comment == "annotationComment"



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

from beartype.typing import List, Optional

from spdx_tools.spdx.model import Document, Snippet
from spdx_tools.spdx.validation.license_expression_validator import (
    validate_license_expression,
    validate_license_expressions,
)
from spdx_tools.spdx.validation.spdx_id_validators import validate_spdx_id
from spdx_tools.spdx.validation.validation_message import SpdxElementType, ValidationContext, ValidationMessage


def validate_snippets(
    snippets: List[Snippet], spdx_version: str, document: Optional[Document] = None
) -> List[ValidationMessage]:
    validation_messages = []
    if document:
        for snippet in snippets:
            validation_messages.extend(validate_snippet_within_document(snippet, spdx_version, document))
    else:
        for snippet in snippets:
            validation_messages.extend(validate_snippet(snippet, spdx_version))

    return validation_messages


def validate_snippet_within_document(
    snippet: Snippet, spdx_version: str, document: Document
) -> List[ValidationMessage]:
    validation_messages: List[ValidationMessage] = []
    context = ValidationContext(
        spdx_id=snippet.spdx_id,
        parent_id=document.creation_info.spdx_id,
        element_type=SpdxElementType.SNIPPET,
        full_element=snippet,
    )

    messages: List[str] = validate_spdx_id(snippet.spdx_id, document)
    for message in messages:
        validation_messages.append(ValidationMessage(message, context))

    messages: List[str] = validate_spdx_id(snippet.file_spdx_id, document, check_files=True)
    for message in messages:
        validation_messages.append(ValidationMessage(message, context))

    validation_messages.extend(validate_license_expression(snippet.license_concluded, document, snippet.spdx_id))

    validation_messages.extend(
        validate_license_expressions(snippet.license_info_in_snippet, document, snippet.spdx_id)
    )

    validation_messages.extend(validate_snippet(snippet, spdx_version, context))

    return validation_messages


def validate_snippet(
    snippet: Snippet, spdx_version: str, context: Optional[ValidationContext] = None
) -> List[ValidationMessage]:
    validation_messages = []
    if not context:
        context = ValidationContext(
            spdx_id=snippet.spdx_id, element_type=SpdxElementType.SNIPPET, full_element=snippet
        )

    if snippet.byte_range[0] < 1:
        validation_messages.append(
            ValidationMessage(
                f"byte_range values must be greater than or equal to 1, but is: {snippet.byte_range}", context
            )
        )

    if snippet.byte_range[0] > snippet.byte_range[1]:
        validation_messages.append(
            ValidationMessage(
                f"the first value of byte_range must be less than or equal to the second, but is: "
                f"{snippet.byte_range}",
                context,
            )
        )

    if snippet.line_range:
        if snippet.line_range[0] < 1:
            validation_messages.append(
                ValidationMessage(
                    f"line_range values must be greater than or equal to 1, but is: {snippet.line_range}", context
                )
            )

        if snippet.line_range[0] > snippet.line_range[1]:
            validation_messages.append(
                ValidationMessage(
                    f"the first value of line_range must be less than or equal to the second, "
                    f"but is: {snippet.line_range}",
                    context,
                )
            )

    if spdx_version == "SPDX-2.2":
        if snippet.license_concluded is None:
            validation_messages.append(ValidationMessage("license_concluded is mandatory in SPDX-2.2", context))
        if snippet.copyright_text is None:
            validation_messages.append(ValidationMessage("copyright_text is mandatory in SPDX-2.2", context))

    return validation_messages



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import re

from beartype.typing import Any, Callable, Dict, Optional
from ply.yacc import YaccProduction

from spdx_tools.spdx.casing_tools import camel_case_to_snake_case
from spdx_tools.spdx.model import (
    Annotation,
    Checksum,
    ChecksumAlgorithm,
    CreationInfo,
    ExtractedLicensingInfo,
    File,
    Package,
    Snippet,
)
from spdx_tools.spdx.parser.error import SPDXParsingError


def grammar_rule(doc):
    # this is a helper method to use decorators for the parsing methods instead of docstrings
    def decorate(func):
        func.__doc__ = doc
        return func

    return decorate


def str_from_text(text: Optional[str]) -> Optional[str]:
    regex = re.compile("<text>((.|\n)+)</text>", re.UNICODE)
    match = regex.match(text)
    if match:
        return match.group(1)
    elif isinstance(text, str):
        return text
    else:
        return None


def parse_checksum(checksum_str: str) -> Checksum:
    # The lexer and the corresponding regex for the token CHECKSUM and EXT_DOC_REF_CHECKSUM ensure that the passed
    # checksum_str is formatted in the way that the following lines of code can't cause an error.
    algorithm, value = checksum_str.split(":")
    algorithm = ChecksumAlgorithm[algorithm.upper().replace("-", "_")]
    value = value.strip()
    checksum = Checksum(algorithm, value)
    return checksum


def set_value(
    parsed_value: YaccProduction,
    dict_to_fill: Dict[str, Any],
    argument_name: Optional[str] = None,
    method_to_apply: Callable = lambda x: x,
):
    if not argument_name:
        argument_name = get_property_name(parsed_value[1])
    if argument_name in dict_to_fill:
        dict_to_fill["logger"].append(f"Multiple values for {parsed_value[1]} found. Line: {parsed_value.lineno(1)}")
        return
    try:
        dict_to_fill[argument_name] = method_to_apply(parsed_value[2])
    except SPDXParsingError as err:
        dict_to_fill["logger"].append(err.get_messages())
    except ValueError as err:
        dict_to_fill["logger"].append(err.args[0])
    except KeyError:
        dict_to_fill["logger"].append(f"Invalid {parsed_value[1]}: {parsed_value[2]}. Line: {parsed_value.lineno(1)}")


def get_property_name(tag: str):
    if tag not in TAG_DATA_MODEL_FIELD.keys():
        return camel_case_to_snake_case(tag)
    return TAG_DATA_MODEL_FIELD[tag][1]


# This dictionary serves as a mapping from a tag to the corresponding class and field in the internal data model.
# This mapping is not complete as we only list the values which can be parsed by a generic method and don't need any
# individual logic.
TAG_DATA_MODEL_FIELD = {
    "SPDXVersion": (CreationInfo, "spdx_version"),
    "DataLicense": (CreationInfo, "data_license"),
    "DocumentName": (CreationInfo, "name"),
    "DocumentComment": (CreationInfo, "document_comment"),
    "DocumentNamespace": (CreationInfo, "document_namespace"),
    "Creator": (CreationInfo, "creator"),
    "Created": (CreationInfo, "created"),
    "CreatorComment": (CreationInfo, "creator_comment"),
    "LicenseListVersion": (CreationInfo, "license_list_version"),
    "ExternalDocumentRef": (CreationInfo, "external_document_refs"),
    "FileName": (File, "name"),
    "FileType": (File, "file_type"),
    "FileChecksum": (File, "checksums"),
    "FileNotice": (File, "notice"),
    "FileCopyrightText": (File, "copyright_text"),
    "LicenseComments": (File, "license_comment"),
    "FileComment": (File, "comment"),
    "LicenseConcluded": (File, "license_concluded"),
    "LicenseDeclared": (File, "license_declared"),
    "PackageName": (Package, "name"),
    "PackageComment": (Package, "comment"),
    "PackageCopyrightText": (Package, "copyright_text"),
    "PackageLicenseComments": (Package, "license_comment"),
    "PackageLicenseDeclared": (Package, "license_declared"),
    "PackageLicenseConcluded": (Package, "license_concluded"),
    "PackageFileName": (Package, "file_name"),
    "PackageVersion": (Package, "version"),
    "PackageDownloadLocation": (Package, "download_location"),
    "PackageSummary": (Package, "summary"),
    "PackageSourceInfo": (Package, "source_info"),
    "PackageSupplier": (Package, "supplier"),
    "PackageOriginator": (Package, "originator"),
    "PackageDescription": (Package, "description"),
    "PackageHomePage": (Package, "homepage"),
    "SnippetSPDXID": (Snippet, "spdx_id"),
    "SnippetFromFileSPDXID": (Snippet, "file_spdx_id"),
    "SnippetName": (Snippet, "name"),
    "SnippetComment": (Snippet, "comment"),
    "SnippetCopyrightText": (Snippet, "copyright_text"),
    "SnippetLicenseComments": (Snippet, "license_comment"),
    "SnippetLicenseConcluded": (Snippet, "license_concluded"),
    "SnippetByteRange": (Snippet, "byte_range"),
    "SnippetLineRange": (Snippet, "line_range"),
    "Annotator": (Annotation, "annotator"),
    "SPDXREF": (Annotation, "spdx_id"),
    "AnnotationComment": (Annotation, "annotation_comment"),
    "LicenseID": (ExtractedLicensingInfo, "license_id"),
    "ExtractedText": (ExtractedLicensingInfo, "extracted_text"),
    "LicenseComment": (ExtractedLicensingInfo, "comment"),
    "LicenseName": (ExtractedLicensingInfo, "license_name"),
}



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from spdx_tools.spdx3.bump_from_spdx2.relationship import bump_relationship, bump_relationships
from spdx_tools.spdx3.model import Relationship, RelationshipCompleteness, RelationshipType
from spdx_tools.spdx3.payload import Payload
from spdx_tools.spdx.model import RelationshipType as Spdx2_RelationshipType
from spdx_tools.spdx.model import SpdxNoAssertion, SpdxNone
from tests.spdx.fixtures import relationship_fixture


def test_relationship_bump():
    spdx2_relationship = relationship_fixture()
    document_namespace = "https://doc.namespace"
    relationship = bump_relationship(spdx2_relationship, document_namespace, 1)

    assert relationship == Relationship(
        f"{document_namespace}#SPDXRef-Relationship-1",
        f"{document_namespace}#{spdx2_relationship.spdx_element_id}",
        RelationshipType.DESCRIBES,
        [f"{document_namespace}#{spdx2_relationship.related_spdx_element_id}"],
        comment=spdx2_relationship.comment,
    )


def test_relationships_bump():
    relationships = [
        relationship_fixture(comment=None),
        relationship_fixture(related_spdx_element_id="SPDXRef-Package", comment=None),
    ]
    payload = Payload()
    document_namespace = "https://doc.namespace"
    bump_relationships(relationships, payload, document_namespace)

    assert payload.get_element(f"{document_namespace}#SPDXRef-Relationship-1") == Relationship(
        f"{document_namespace}#SPDXRef-Relationship-1",
        f"{document_namespace}#{relationships[0].spdx_element_id}",
        RelationshipType.DESCRIBES,
        [
            f"{document_namespace}#{relationships[0].related_spdx_element_id}",
            f"{document_namespace}#{relationships[1].related_spdx_element_id}",
        ],
    )


def test_relationships_bump_with_setting_completeness():
    relationships = [
        relationship_fixture(related_spdx_element_id=SpdxNoAssertion()),
        relationship_fixture(related_spdx_element_id="SPDXRef-Package"),
        relationship_fixture(
            relationship_type=Spdx2_RelationshipType.SPECIFICATION_FOR,
            related_spdx_element_id=SpdxNone(),
            comment=None,
        ),
    ]
    payload = Payload()
    document_namespace = "https://doc.namespace"
    bump_relationships(relationships, payload, document_namespace)

    assert payload.get_element(f"{document_namespace}#SPDXRef-Relationship-0") == Relationship(
        f"{document_namespace}#SPDXRef-Relationship-0",
        f"{document_namespace}#{relationships[0].spdx_element_id}",
        RelationshipType.DESCRIBES,
        [],
        comment=relationships[0].comment,
        completeness=RelationshipCompleteness.NOASSERTION,
    )
    assert payload.get_element(f"{document_namespace}#SPDXRef-Relationship-1") == Relationship(
        f"{document_namespace}#SPDXRef-Relationship-1",
        f"{document_namespace}#{relationships[1].spdx_element_id}",
        RelationshipType.DESCRIBES,
        [f"{document_namespace}#{relationships[1].related_spdx_element_id}"],
        comment=relationships[1].comment,
    )
    assert payload.get_element(f"{document_namespace}#SPDXRef-Relationship-2") == Relationship(
        f"{document_namespace}#SPDXRef-Relationship-2",
        f"{document_namespace}#{relationships[2].spdx_element_id}",
        RelationshipType.SPECIFICATION_FOR,
        [],
        completeness=RelationshipCompleteness.COMPLETE,
    )


def test_undefined_relationship_bump(capsys):
    relationships = [
        relationship_fixture(
            related_spdx_element_id=SpdxNoAssertion(), relationship_type=Spdx2_RelationshipType.CONTAINED_BY
        ),
        relationship_fixture(relationship_type=Spdx2_RelationshipType.OPTIONAL_COMPONENT_OF),
    ]
    payload = Payload()
    document_namespace = "https://doc.namespace"
    bump_relationships(relationships, payload, document_namespace)

    captured = capsys.readouterr()
    assert (
        captured.err == "Swapped Relationship to NoAssertion/None not converted: missing conversion rule \n"
        "OPTIONAL_COMPONENT_OF not converted: missing conversion rule \n"
    )



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from pathlib import Path
from typing import List
from unittest import TestCase

import pytest

from spdx_tools.spdx.graph_generation import generate_relationship_graph_from_spdx
from spdx_tools.spdx.model import Document, Relationship, RelationshipType
from spdx_tools.spdx.parser.parse_anything import parse_file
from tests.spdx.fixtures import document_fixture, file_fixture, package_fixture

try:
    import networkx  # noqa: F401
except ImportError:
    pytest.skip("Skip this module as the tests need optional dependencies to run.", allow_module_level=True)


@pytest.mark.parametrize(
    "file_name, nodes_count, edges_count, relationship_node_keys",
    [
        (
            "SPDXJSONExample-v2.3.spdx.json",
            22,
            22,
            ["SPDXRef-Package_DYNAMIC_LINK", "SPDXRef-JenaLib_CONTAINS"],
        ),
        (
            "SPDXJSONExample-v2.2.spdx.json",
            20,
            19,
            ["SPDXRef-Package_DYNAMIC_LINK", "SPDXRef-JenaLib_CONTAINS"],
        ),
        (
            "SPDXRdfExample-v2.3.spdx.rdf.xml",
            22,
            22,
            ["SPDXRef-Package_DYNAMIC_LINK", "SPDXRef-JenaLib_CONTAINS"],
        ),
        (
            "SPDXRdfExample-v2.2.spdx.rdf.xml",
            20,
            19,
            ["SPDXRef-Package_DYNAMIC_LINK", "SPDXRef-JenaLib_CONTAINS"],
        ),
        (
            "SPDXTagExample-v2.3.spdx",
            22,
            22,
            ["SPDXRef-Package_DYNAMIC_LINK", "SPDXRef-JenaLib_CONTAINS"],
        ),
    ],
)
def test_generate_graph_from_spdx(
    file_name: str,
    nodes_count: int,
    edges_count: int,
    relationship_node_keys: List[str],
) -> None:
    document = parse_file(str(Path(__file__).resolve().parent.parent / "spdx" / "data" / file_name))
    graph = generate_relationship_graph_from_spdx(document)

    assert document.creation_info.spdx_id in graph.nodes()
    assert graph.number_of_nodes() == nodes_count
    assert graph.number_of_edges() == edges_count
    assert "SPDXRef-DOCUMENT_DESCRIBES" in graph.nodes()
    for relationship_node_key in relationship_node_keys:
        assert relationship_node_key in graph.nodes()


def test_complete_connected_graph() -> None:
    document = _create_minimal_document()

    graph = generate_relationship_graph_from_spdx(document)

    TestCase().assertCountEqual(
        graph.nodes(),
        [
            "SPDXRef-DOCUMENT",
            "SPDXRef-Package-A",
            "SPDXRef-Package-B",
            "SPDXRef-File",
            "SPDXRef-DOCUMENT_DESCRIBES",
            "SPDXRef-Package-A_CONTAINS",
            "SPDXRef-Package-B_CONTAINS",
        ],
    )
    TestCase().assertCountEqual(
        graph.edges(),
        [
            ("SPDXRef-DOCUMENT", "SPDXRef-DOCUMENT_DESCRIBES"),
            ("SPDXRef-DOCUMENT_DESCRIBES", "SPDXRef-Package-A"),
            ("SPDXRef-DOCUMENT_DESCRIBES", "SPDXRef-Package-B"),
            ("SPDXRef-Package-A", "SPDXRef-Package-A_CONTAINS"),
            ("SPDXRef-Package-A_CONTAINS", "SPDXRef-File"),
            ("SPDXRef-Package-B", "SPDXRef-Package-B_CONTAINS"),
            ("SPDXRef-Package-B_CONTAINS", "SPDXRef-File"),
        ],
    )


def test_complete_unconnected_graph() -> None:
    document = _create_minimal_document()
    document.packages += [package_fixture(spdx_id="SPDXRef-Package-C", name="Package without connection to document")]

    graph = generate_relationship_graph_from_spdx(document)

    TestCase().assertCountEqual(
        graph.nodes(),
        [
            "SPDXRef-DOCUMENT",
            "SPDXRef-Package-A",
            "SPDXRef-Package-B",
            "SPDXRef-File",
            "SPDXRef-DOCUMENT_DESCRIBES",
            "SPDXRef-Package-A_CONTAINS",
            "SPDXRef-Package-B_CONTAINS",
            "SPDXRef-Package-C",
        ],
    )
    TestCase().assertCountEqual(
        graph.edges(),
        [
            ("SPDXRef-DOCUMENT", "SPDXRef-DOCUMENT_DESCRIBES"),
            ("SPDXRef-DOCUMENT_DESCRIBES", "SPDXRef-Package-A"),
            ("SPDXRef-DOCUMENT_DESCRIBES", "SPDXRef-Package-B"),
            ("SPDXRef-Package-A", "SPDXRef-Package-A_CONTAINS"),
            ("SPDXRef-Package-A_CONTAINS", "SPDXRef-File"),
            ("SPDXRef-Package-B", "SPDXRef-Package-B_CONTAINS"),
            ("SPDXRef-Package-B_CONTAINS", "SPDXRef-File"),
        ],
    )


def _create_minimal_document() -> Document:
    packages = [
        package_fixture(spdx_id="SPDXRef-Package-A", name="Package-A"),
        package_fixture(spdx_id="SPDXRef-Package-B", name="Package-B"),
    ]
    files = [
        file_fixture(spdx_id="SPDXRef-File", name="File"),
    ]
    relationships = [
        Relationship("SPDXRef-DOCUMENT", RelationshipType.DESCRIBES, "SPDXRef-Package-A"),
        Relationship("SPDXRef-DOCUMENT", RelationshipType.DESCRIBES, "SPDXRef-Package-B"),
        Relationship("SPDXRef-Package-A", RelationshipType.CONTAINS, "SPDXRef-File"),
        Relationship("SPDXRef-Package-B", RelationshipType.CONTAINS, "SPDXRef-File"),
    ]
    document = document_fixture(packages=packages, files=files, relationships=relationships, snippets=[])

    return document



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from enum import auto

from spdx_tools.spdx.jsonschema.json_property import JsonProperty


class SnippetProperty(JsonProperty):
    SPDX_ID = auto()
    ANNOTATIONS = auto()
    ATTRIBUTION_TEXTS = auto()
    COMMENT = auto()
    COPYRIGHT_TEXT = auto()
    LICENSE_COMMENTS = auto()
    LICENSE_CONCLUDED = auto()
    LICENSE_INFO_IN_SNIPPETS = auto()
    NAME = auto()
    RANGES = auto()
    SNIPPET_FROM_FILE = auto()



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

from beartype.typing import List, Optional, Union
from license_expression import ExpressionError, ExpressionParseError, LicenseExpression, get_spdx_licensing

from spdx_tools.spdx.model import Document, SpdxNoAssertion, SpdxNone
from spdx_tools.spdx.validation.validation_message import SpdxElementType, ValidationContext, ValidationMessage


def validate_license_expressions(
    license_expressions: List[Union[LicenseExpression, SpdxNoAssertion, SpdxNone]], document: Document, parent_id: str
) -> List[ValidationMessage]:
    context = ValidationContext(
        parent_id=parent_id, element_type=SpdxElementType.LICENSE_EXPRESSION, full_element=license_expressions
    )
    validation_messages = []

    for license_expression in license_expressions:
        validation_messages.extend(validate_license_expression(license_expression, document, parent_id, context))

    return validation_messages


def validate_license_expression(
    license_expression: Optional[Union[LicenseExpression, SpdxNoAssertion, SpdxNone]],
    document: Document,
    parent_id: str,
    context: ValidationContext = None,
) -> List[ValidationMessage]:
    if license_expression in [SpdxNoAssertion(), SpdxNone(), None]:
        return []

    if not context:
        context = ValidationContext(
            parent_id=parent_id, element_type=SpdxElementType.LICENSE_EXPRESSION, full_element=license_expression
        )

    validation_messages = []
    license_ref_ids: List[str] = [license_ref.license_id for license_ref in document.extracted_licensing_info]

    for non_spdx_token in get_spdx_licensing().validate(license_expression).invalid_symbols:
        if non_spdx_token not in license_ref_ids:
            validation_messages.append(
                ValidationMessage(
                    f"Unrecognized license reference: {non_spdx_token}. license_expression must only use IDs from the "
                    f"license list or extracted licensing info, but is: {license_expression}",
                    context,
                )
            )

    try:
        get_spdx_licensing().parse(str(license_expression), validate=True, strict=True)
    except ExpressionParseError as err:
        # This error is raised when an exception symbol is used as a license symbol and vice versa.
        # So far, it only catches the first such error in the provided string.
        validation_messages.append(ValidationMessage(f"{err}. for license_expression: {license_expression}", context))
    except ExpressionError:
        # This error is raised for invalid symbols within the license_expression, but it provides only a string of
        # these. On the other hand, get_spdx_licensing().validate() gives an actual list of invalid symbols, so this is
        # handled above.
        pass

    return validation_messages



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from datetime import datetime
from typing import Any, Dict, Type

from semantic_version import Version

from spdx_tools.spdx3.model import (
    Agent,
    Annotation,
    AnnotationType,
    Bom,
    Bundle,
    CreationInfo,
    ExternalIdentifier,
    ExternalIdentifierType,
    ExternalMap,
    ExternalReference,
    ExternalReferenceType,
    Hash,
    HashAlgorithm,
    LifecycleScopedRelationship,
    LifecycleScopeType,
    NamespaceMap,
    Organization,
    Person,
    ProfileIdentifier,
    Relationship,
    RelationshipCompleteness,
    RelationshipType,
    SoftwareAgent,
    SpdxDocument,
    Tool,
)
from spdx_tools.spdx3.model.ai.ai_package import AIPackage, SafetyRiskAssessmentType
from spdx_tools.spdx3.model.build import Build
from spdx_tools.spdx3.model.dataset.dataset import (
    ConfidentialityLevelType,
    Dataset,
    DatasetAvailabilityType,
    DatasetType,
)
from spdx_tools.spdx3.model.licensing import (
    CustomLicense,
    CustomLicenseAddition,
    ListedLicense,
    ListedLicenseException,
)
from spdx_tools.spdx3.model.positive_integer_range import PositiveIntegerRange
from spdx_tools.spdx3.model.security import (
    CvssV2VulnAssessmentRelationship,
    CvssV3VulnAssessmentRelationship,
    EpssVulnAssessmentRelationship,
    ExploitCatalogType,
    ExploitCatalogVulnAssessmentRelationship,
    SsvcDecisionType,
    SsvcVulnAssessmentRelationship,
    VexAffectedVulnAssessmentRelationship,
    VexFixedVulnAssessmentRelationship,
    VexJustificationType,
    VexNotAffectedVulnAssessmentRelationship,
    VexUnderInvestigationVulnAssessmentRelationship,
    Vulnerability,
)
from spdx_tools.spdx3.model.software import (
    DependencyConditionalityType,
    File,
    Package,
    Sbom,
    SBOMType,
    Snippet,
    SoftwareDependencyLinkType,
    SoftwareDependencyRelationship,
    SoftwarePurpose,
)

"""Utility methods to create data model instances. All properties have valid defaults, so they don't need to be
specified unless relevant for the test."""


def creation_info_fixture(
    spec_version=Version("3.0.0"),
    created=datetime(2022, 12, 1),
    created_by=None,
    created_using=None,
    profile=None,
    data_license="CC0-1.0",
    comment="creationInfoComment",
) -> CreationInfo:
    created_by = ["https://spdx.test/tools-python/creation_info_created_by"] if created_by is None else created_by
    created_using = (
        ["https://spdx.test/tools-python/creation_info_created_using"] if created_using is None else created_using
    )
    profile = (
        [ProfileIdentifier.CORE, ProfileIdentifier.SOFTWARE, ProfileIdentifier.LICENSING]
        if profile is None
        else profile
    )
    return CreationInfo(
        spec_version=spec_version,
        created=created,
        created_by=created_by,
        profile=profile,
        data_license=data_license,
        created_using=created_using,
        comment=comment,
    )


def external_identifier_fixture(
    external_identifier_type=ExternalIdentifierType.OTHER,
    identifier="externalIdentifierIdentifier",
    comment="externalIdentifierComment",
    identifier_locator=None,
    issuing_authority="https://spdx.test/tools-python/external_identifier_issuing_authority",
) -> ExternalIdentifier:
    identifier_locator = (
        ["https://spdx.test/tools-python/external_identifier_identifier_locator"]
        if identifier_locator is None
        else identifier_locator
    )
    return ExternalIdentifier(
        external_identifier_type=external_identifier_type,
        identifier=identifier,
        comment=comment,
        identifier_locator=identifier_locator,
        issuing_authority=issuing_authority,
    )


def external_reference_fixture(
    external_reference_type=ExternalReferenceType.OTHER,
    locator=None,
    content_type="externalReferenceContentType",
    comment="externalReferenceComment",
) -> ExternalReference:
    locator = ["org.apache.tomcat:tomcat:9.0.0.M4"] if locator is None else locator
    return ExternalReference(
        external_reference_type=external_reference_type, locator=locator, content_type=content_type, comment=comment
    )


def hash_fixture(
    algorithm=HashAlgorithm.SHA1,
    hash_value="71c4025dd9897b364f3ebbb42c484ff43d00791c",
    comment="hashComment",
) -> Hash:
    return Hash(algorithm=algorithm, hash_value=hash_value, comment=comment)


def external_map_fixture(
    external_id="https://spdx.test/tools-python/external_map_external_id",
    verified_using=None,
    location_hint="https://spdx.test/tools-python/external_map_location_hint",
    defining_document="https://spdx.test/tools-python/defining_document",
) -> ExternalMap:
    verified_using = [hash_fixture()] if verified_using is None else verified_using
    return ExternalMap(
        external_id=external_id,
        verified_using=verified_using,
        location_hint=location_hint,
        defining_document=defining_document,
    )


def namespace_map_fixture(
    prefix="namespaceMapPrefix", namespace="https://spdx.test/tools-python/namespace_map_namespace"
) -> NamespaceMap:
    return NamespaceMap(prefix=prefix, namespace=namespace)


def listed_license_fixture(
    license_id="https://spdx.test/tools-python/license_id",
    license_name="license name",
    license_text="license text",
    license_comment="license comment",
    see_also=None,
    is_osi_approved=True,
    is_fsf_libre=True,
    standard_license_header="license header",
    standard_license_template="license template",
    is_deprecated_license_id=True,
    obsoleted_by="https://spdx.test/tools-python/obsoleted_by_license_id",
    list_version_added="2.1",
    deprecated_version="2.2",
):
    see_also = ["https://see.also/license"] if see_also is None else see_also
    return ListedLicense(
        license_id=license_id,
        license_name=license_name,
        license_text=license_text,
        license_comment=license_comment,
        see_also=see_also,
        is_osi_approved=is_osi_approved,
        is_fsf_libre=is_fsf_libre,
        standard_license_header=standard_license_header,
        standard_license_template=standard_license_template,
        is_deprecated_license_id=is_deprecated_license_id,
        obsoleted_by=obsoleted_by,
        list_version_added=list_version_added,
        deprecated_version=deprecated_version,
    )


ELEMENT_DICT = {
    "spdx_id": "https://spdx.test/tools-python/element_fixture",
    "creation_info": creation_info_fixture(),
    "name": "elementName",
    "summary": "elementSummary",
    "description": "elementDescription",
    "comment": "elementComment",
    "verified_using": [hash_fixture()],
    "external_reference": [external_reference_fixture()],
    "external_identifier": [external_identifier_fixture()],
    "extension": "extensionPlaceholder",
}

RELATIONSHIP_DICT = {
    "from_element": "https://spdx.test/tools-python/relationship_from_element",
    "relationship_type": RelationshipType.OTHER,
    "to": ["https://spdx.test/tools-python/relationship_to"],
    "completeness": RelationshipCompleteness.COMPLETE,
    "start_time": datetime(2020, 1, 1),
    "end_time": datetime(2023, 1, 1),
}

LIFECYCLE_SCOPED_RELATIONSHIP_DICT = {"scope": LifecycleScopeType.DESIGN}

ANNOTATION_DICT = {
    "annotation_type": AnnotationType.OTHER,
    "subject": "https://spdx.test/tools-python/annotation_subject",
    "content_type": ["annotationContent"],
    "statement": "annotationStatement",
}

ELEMENT_COLLECTION_DICT = {
    "element": ["https://spdx.test/tools-python/collection_element"],
    "root_element": ["https://spdx.test/tools-python/collection_root_element"],
    "namespaces": [namespace_map_fixture()],
    "imports": [external_map_fixture()],
}

BUNDLE_DICT = {
    "context": "bundleContext",
}

SBOM_DICT = {
    "sbom_type": [SBOMType.BUILD],
}

LICENSE_DICT = {
    "license_id": "https://spdx.test/tools-python/license_id",
    "license_name": "license name",
    "license_text": "license text",
    "license_comment": "license comment",
    "see_also": ["https://see.also/license"],
    "is_osi_approved": True,
    "is_fsf_libre": True,
    "standard_license_header": "license header",
    "standard_license_template": "license template",
    "is_deprecated_license_id": True,
    "obsoleted_by": "https://spdx.test/tools-python/obsoleted_by_license_id",
}

LISTED_LICENSE_DICT = {
    "list_version_added": "2.1",
    "deprecated_version": "2.2",
}

LICENSE_ADDITION_DICT = {
    "addition_id": "https://spdx.test/tools-python/addition_id",
    "addition_name": "addition name",
    "addition_text": "addition text",
    "addition_comment": "addition comment",
    "see_also": ["https://see.also/addition"],
    "standard_addition_template": "addition template",
    "is_deprecated_addition_id": True,
    "obsoleted_by": "https://spdx.test/tools-python/obsoleted_by_addition_id",
}

LISTED_LICENSE_EXCEPTION_DICT = {
    "list_version_added": "2.1",
    "deprecated_version": "2.2",
}

VULNERABILITY_DICT = {
    "published_time": datetime(2010, 1, 1),
    "modified_time": datetime(2011, 1, 1),
    "withdrawn_time": datetime(2012, 1, 1),
}

VULN_ASSESSMENT_RELATIONSHIP_DICT = {
    "assessed_element": "https://spdx.test/tools-python/assessed_element",
    "published_time": datetime(2004, 1, 1),
    "supplied_by": "https://spdx.test/tools-python/supplied_by",
    "modified_time": datetime(2005, 1, 1),
    "withdrawn_time": datetime(2006, 1, 1),
}

CVSS_V2_VULN_ASSESSMENT_RELATIONSHIP_DICT = {
    "score": "4.3",
    "severity": "low",
    "vector": "(AV:N/AC:M/Au:N/C:P/I:N/A:N)",
    "relationship_type": RelationshipType.HAS_ASSESSMENT_FOR,
}

CVSS_V3_VULN_ASSESSMENT_RELATIONSHIP_DICT = {
    "score": "6.8",
    "severity": "medium",
    "vector": "CVSS:3.1/AV:N/AC:H/PR:N/UI:N/S:C/C:H/I:N/A:N",
    "relationship_type": RelationshipType.HAS_ASSESSMENT_FOR,
}

EPSS_VULN_ASSESSMENT_RELATIONSHIP_DICT = {
    "probability": 80,
    "severity": "high",
    "relationship_type": RelationshipType.HAS_ASSESSMENT_FOR,
}

SSVC_VULN_ASSESSMENT_RELATIONSHIP_DICT = {
    "decision_type": SsvcDecisionType.ACT,
    "relationship_type": RelationshipType.HAS_ASSESSMENT_FOR,
}

EXPLOIT_CATALOG_VULN_ASSESSMENT_RELATIONSHIP_DICT = {
    "catalog_type": ExploitCatalogType.KEV,
    "exploited": True,
    "locator": "https://www.cisa.gov/known-exploited-vulnerabilities-catalog",
    "relationship_type": RelationshipType.HAS_ASSESSMENT_FOR,
}

VEX_VULN_ASSESSMENT_RELATIONSHIP_DICT = {
    "vex_version": "v4.2",
    "status_notes": "some status notes",
}

VEX_AFFECTED_VULN_ASSESSMENT_RELATIONSHIP_DICT = {
    "action_statement": "Upgrade to version 1.4 of ACME application.",
    "action_statement_time": [datetime(2015, 10, 15)],
    "relationship_type": RelationshipType.AFFECTS,
}

VEX_NOT_AFFECTED_VULN_ASSESSMENT_RELATIONSHIP_DICT = {
    "justification_type": VexJustificationType.COMPONENT_NOT_PRESENT,
    "impact_statement": "Not using this vulnerable part of this library.",
    "impact_statement_time": datetime(2015, 10, 15),
    "relationship_type": RelationshipType.DOES_NOT_AFFECT,
}

VEX_UNDER_INVESTIGATION_VULN_ASSESSMENT_RELATIONSHIP_DICT = {
    "relationship_type": RelationshipType.UNDER_INVESTIGATION_FOR,
}

VEX_FIXED_VULN_ASSESSMENT_RELATIONSHIP_DICT = {
    "relationship_type": RelationshipType.FIXED_IN,
}

AIPACKAGE_DICT = {
    "energy_consumption": "energyConsumption",
    "standard_compliance": ["standardCompliance"],
    "limitation": "aIPackageLimitation",
    "type_of_model": ["typeOfModel"],
    "information_about_training": "informationAboutTraining",
    "information_about_application": "informationAboutApplication",
    "hyperparameter": {"aIPackageHypParaKey": "aIPackageHypParaValue"},
    "model_data_preprocessing": ["aImodelDataPreprocessing"],
    "model_explainability": ["aImodelExplainability"],
    "sensitive_personal_information": True,
    "metric_decision_threshold": {"metricDecisionThresholdKey": "metricDecisionThresholdValue"},
    "metric": {"aIMetricKey": "aIMetricValue"},
    "domain": ["aIDomain"],
    "autonomy_type": True,
    "safety_risk_assessment": SafetyRiskAssessmentType.LOW,
}

ARTIFACT_DICT = {
    "originated_by": ["https://spdx.test/tools-python/originatedBy"],
    "supplied_by": ["https://spdx.test/tools-python/suppliedBy"],
    "built_time": datetime(2004, 1, 1),
    "release_time": datetime(2005, 1, 1),
    "valid_until_time": datetime(2006, 1, 1),
    "standard": ["https://spdx.test/tools-python/standard"],
}

SOFTWARE_ARTIFACT_DICT = {
    "content_identifier": "https://spdx.test/tools-python/contentIdentifier",
    "primary_purpose": SoftwarePurpose.SOURCE,
    "additional_purpose": [SoftwarePurpose.OTHER],
    "concluded_license": listed_license_fixture(),
    "declared_license": listed_license_fixture(),
    "copyright_text": "copyrightText",
    "attribution_text": "attributionText",
}

FILE_DICT = {"content_type": "fileContentType"}

PACKAGE_DICT = {
    "package_version": "packageVersion",
    "download_location": "https://spdx.test/tools-python/downloadPackage",
    "package_url": "https://spdx.test/tools-python/package",
    "homepage": "https://spdx.test/tools-python/homepage",
    "source_info": "sourceInfo",
}

SNIPPET_DICT = {"byte_range": PositiveIntegerRange(1024, 2048), "line_range": PositiveIntegerRange(1, 4)}

SOFTWARE_DEPENDENCY_RELATIONSHIP_DICT = {
    "software_linkage": SoftwareDependencyLinkType.OTHER,
    "conditionality": DependencyConditionalityType.OTHER,
}

DATASET_DICT = {
    "dataset_type": [DatasetType.OTHER],
    "data_collection_process": "DatasetDataCollectionProcess",
    "intended_use": "DatasetIntendedUse",
    "dataset_size": 10,
    "dataset_noise": "DatasetNoise",
    "data_preprocessing": ["DataPreprocessing"],
    "sensor": {"SensorKey": "SensorValue"},
    "known_bias": ["DatasetKnownBias"],
    "sensitive_personal_information": True,
    "anonymization_method_used": ["DatasetAnonymizationMethodUsed"],
    "confidentiality_level": ConfidentialityLevelType.CLEAR,
    "dataset_update_mechanism": "DatasetUpdateMechanism",
    "dataset_availability": DatasetAvailabilityType.QUERY,
}

BUILD_DICT = {
    "build_type": "BuildType",
    "build_id": "BuildId",
    "config_source_entrypoint": ["ConfigSourceEntrypoint"],
    "config_source_uri": ["ConfigSourceURI"],
    "config_source_digest": [hash_fixture()],
    "parameters": {"parameter": "value"},
    "build_start_time": datetime(2015, 4, 4),
    "build_end_time": datetime(2015, 4, 5),
    "environment": {"environment_param": "environment_value"},
}

FIXTURE_DICTS = {
    Agent: [ELEMENT_DICT],
    Person: [ELEMENT_DICT],
    Organization: [ELEMENT_DICT],
    SoftwareAgent: [ELEMENT_DICT],
    Tool: [ELEMENT_DICT],
    Relationship: [ELEMENT_DICT, RELATIONSHIP_DICT],
    LifecycleScopedRelationship: [ELEMENT_DICT, RELATIONSHIP_DICT, LIFECYCLE_SCOPED_RELATIONSHIP_DICT],
    Annotation: [ELEMENT_DICT, ANNOTATION_DICT],
    Bundle: [ELEMENT_DICT, ELEMENT_COLLECTION_DICT, BUNDLE_DICT],
    SpdxDocument: [ELEMENT_DICT, ELEMENT_COLLECTION_DICT, BUNDLE_DICT],
    Bom: [ELEMENT_DICT, ELEMENT_COLLECTION_DICT, BUNDLE_DICT],
    Sbom: [ELEMENT_DICT, ELEMENT_COLLECTION_DICT, BUNDLE_DICT, SBOM_DICT],
    ListedLicense: [LICENSE_DICT, LISTED_LICENSE_DICT],
    CustomLicense: [LICENSE_DICT],
    ListedLicenseException: [LICENSE_ADDITION_DICT, LISTED_LICENSE_EXCEPTION_DICT],
    CustomLicenseAddition: [LICENSE_ADDITION_DICT],
    CvssV2VulnAssessmentRelationship: [
        ELEMENT_DICT,
        RELATIONSHIP_DICT,
        VULN_ASSESSMENT_RELATIONSHIP_DICT,
        CVSS_V2_VULN_ASSESSMENT_RELATIONSHIP_DICT,
    ],
    CvssV3VulnAssessmentRelationship: [
        ELEMENT_DICT,
        RELATIONSHIP_DICT,
        VULN_ASSESSMENT_RELATIONSHIP_DICT,
        CVSS_V3_VULN_ASSESSMENT_RELATIONSHIP_DICT,
    ],
    EpssVulnAssessmentRelationship: [
        ELEMENT_DICT,
        RELATIONSHIP_DICT,
        VULN_ASSESSMENT_RELATIONSHIP_DICT,
        EPSS_VULN_ASSESSMENT_RELATIONSHIP_DICT,
    ],
    ExploitCatalogVulnAssessmentRelationship: [
        ELEMENT_DICT,
        RELATIONSHIP_DICT,
        VULN_ASSESSMENT_RELATIONSHIP_DICT,
        EXPLOIT_CATALOG_VULN_ASSESSMENT_RELATIONSHIP_DICT,
    ],
    SsvcVulnAssessmentRelationship: [
        ELEMENT_DICT,
        RELATIONSHIP_DICT,
        VULN_ASSESSMENT_RELATIONSHIP_DICT,
        SSVC_VULN_ASSESSMENT_RELATIONSHIP_DICT,
    ],
    VexAffectedVulnAssessmentRelationship: [
        ELEMENT_DICT,
        RELATIONSHIP_DICT,
        VULN_ASSESSMENT_RELATIONSHIP_DICT,
        VEX_VULN_ASSESSMENT_RELATIONSHIP_DICT,
        VEX_AFFECTED_VULN_ASSESSMENT_RELATIONSHIP_DICT,
    ],
    VexFixedVulnAssessmentRelationship: [
        ELEMENT_DICT,
        RELATIONSHIP_DICT,
        VULN_ASSESSMENT_RELATIONSHIP_DICT,
        VEX_VULN_ASSESSMENT_RELATIONSHIP_DICT,
        VEX_FIXED_VULN_ASSESSMENT_RELATIONSHIP_DICT,
    ],
    VexNotAffectedVulnAssessmentRelationship: [
        ELEMENT_DICT,
        RELATIONSHIP_DICT,
        VULN_ASSESSMENT_RELATIONSHIP_DICT,
        VEX_VULN_ASSESSMENT_RELATIONSHIP_DICT,
        VEX_NOT_AFFECTED_VULN_ASSESSMENT_RELATIONSHIP_DICT,
    ],
    VexUnderInvestigationVulnAssessmentRelationship: [
        ELEMENT_DICT,
        RELATIONSHIP_DICT,
        VULN_ASSESSMENT_RELATIONSHIP_DICT,
        VEX_VULN_ASSESSMENT_RELATIONSHIP_DICT,
        VEX_UNDER_INVESTIGATION_VULN_ASSESSMENT_RELATIONSHIP_DICT,
    ],
    Vulnerability: [ELEMENT_DICT, VULNERABILITY_DICT],
    AIPackage: [AIPACKAGE_DICT, PACKAGE_DICT, ELEMENT_DICT, ARTIFACT_DICT, SOFTWARE_ARTIFACT_DICT],
    File: [ELEMENT_DICT, ARTIFACT_DICT, SOFTWARE_ARTIFACT_DICT, FILE_DICT],
    Package: [ELEMENT_DICT, ARTIFACT_DICT, SOFTWARE_ARTIFACT_DICT, PACKAGE_DICT],
    Snippet: [ELEMENT_DICT, ARTIFACT_DICT, SOFTWARE_ARTIFACT_DICT, SNIPPET_DICT],
    SoftwareDependencyRelationship: [
        ELEMENT_DICT,
        RELATIONSHIP_DICT,
        LIFECYCLE_SCOPED_RELATIONSHIP_DICT,
        SOFTWARE_DEPENDENCY_RELATIONSHIP_DICT,
    ],
    Dataset: [ELEMENT_DICT, ARTIFACT_DICT, SOFTWARE_ARTIFACT_DICT, PACKAGE_DICT, DATASET_DICT],
    Build: [ELEMENT_DICT, BUILD_DICT],
}


def fixture_factory(clazz: Type[Any], **kwargs) -> Any:
    fixture_dict = get_fixture_dict(clazz)

    for key in kwargs.keys():
        if key not in fixture_dict.keys():
            raise ValueError(f"Provided property name {key} is not part of {clazz.__name__}.")
        else:
            fixture_dict[key] = kwargs[key]

    return clazz(**fixture_dict)


def get_fixture_dict(clazz: Type[Any]) -> Dict[str, Any]:
    fixture_dict = {}
    if clazz not in FIXTURE_DICTS.keys():
        raise ValueError(
            f"{clazz.__name__} is not part of the FIXTURE_DICTS. "
            f"If it is a non-abstract subclass of Element, it was probably forgotten to add."
        )
    for property_dict in FIXTURE_DICTS[clazz]:
        fixture_dict.update(property_dict)

    if "spdx_id" in fixture_dict.keys():
        fixture_dict["spdx_id"] = f"https://spdx.test/tools-python/{clazz.__name__}_fixture"

    return fixture_dict



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

import pytest

from spdx_tools.spdx.model import PackageVerificationCode
from spdx_tools.spdx.validation.package_verification_code_validator import validate_verification_code
from spdx_tools.spdx.validation.validation_message import SpdxElementType, ValidationContext, ValidationMessage


def test_valid_package_verification_code():
    code = PackageVerificationCode("71c4025dd9897b364f3ebbb42c484ff43d00791c", ["./excluded_file", "another.file"])
    validation_messages = validate_verification_code(code, "SPDXRef-Package")

    assert validation_messages == []


@pytest.mark.parametrize(
    "code, expected_message",
    [
        (
            PackageVerificationCode("71c4025dd9897b364f3ebbb42c484ff43d00791cab", []),
            "value of verification_code must consist of 40 lowercase hexadecimal digits, but is: "
            "71c4025dd9897b364f3ebbb42c484ff43d00791cab (length: 42 digits)",
        ),
        (
            PackageVerificationCode("CE9F343C4BA371746FD7EAD9B59031AE34D8AFC4", []),
            "value of verification_code must consist of 40 lowercase hexadecimal digits, but is: "
            "CE9F343C4BA371746FD7EAD9B59031AE34D8AFC4 (length: 40 digits)",
        ),
        (
            PackageVerificationCode("71c4025dd9897b364f3ebbb42c484ff43d00791c", ["/invalid/excluded/file"]),
            'file name must not be an absolute path starting with "/", but is: /invalid/excluded/file',
        ),
    ],
)
def test_invalid_package_verification_code(code, expected_message):
    parent_id = "SPDXRef-Package"
    context = ValidationContext(
        parent_id=parent_id, element_type=SpdxElementType.PACKAGE_VERIFICATION_CODE, full_element=code
    )
    validation_messages = validate_verification_code(code, parent_id)

    assert validation_messages == [ValidationMessage(expected_message, context)]



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import os

import pytest
from rdflib import RDF, Graph, URIRef

from spdx_tools.spdx.constants import DOCUMENT_SPDX_ID
from spdx_tools.spdx.model import RelationshipType
from spdx_tools.spdx.parser.rdf.relationship_parser import parse_implicit_relationship, parse_relationship
from spdx_tools.spdx.rdfschema.namespace import SPDX_NAMESPACE


def test_relationship_parser():
    graph = Graph().parse(os.path.join(os.path.dirname(__file__), "data/file_to_test_rdf_parser.rdf.xml"))
    parent_node = graph.value(predicate=RDF.type, object=SPDX_NAMESPACE.SpdxDocument)
    relationship_node = graph.value(subject=parent_node, predicate=SPDX_NAMESPACE.relationship)
    doc_namespace = "https://some.namespace"
    assert isinstance(parent_node, URIRef)

    relationship = parse_relationship(relationship_node, graph, parent_node, doc_namespace)

    assert relationship.spdx_element_id == DOCUMENT_SPDX_ID
    assert relationship.relationship_type == RelationshipType.DESCRIBES
    assert relationship.related_spdx_element_id == "SPDXRef-File"
    assert relationship.comment == "relationshipComment"


@pytest.mark.parametrize(
    "parent_node, predicate, spdx_element_id, relationship_type, related_spdx_element_id",
    [
        (
            SPDX_NAMESPACE.SpdxDocument,
            SPDX_NAMESPACE.describesPackage,
            DOCUMENT_SPDX_ID,
            RelationshipType.DESCRIBES,
            "SPDXRef-Package",
        ),
        (SPDX_NAMESPACE.Package, SPDX_NAMESPACE.hasFile, "SPDXRef-Package", RelationshipType.CONTAINS, "SPDXRef-File"),
    ],
)
def test_parse_implicit_relationship(
    parent_node, predicate, spdx_element_id, relationship_type, related_spdx_element_id
):
    graph = Graph().parse(os.path.join(os.path.dirname(__file__), "data/file_to_test_rdf_parser.rdf.xml"))
    parent_node = graph.value(predicate=RDF.type, object=parent_node)
    relationship_node = graph.value(subject=parent_node, predicate=predicate)
    assert isinstance(relationship_node, URIRef)
    assert isinstance(parent_node, URIRef)

    doc_namespace = "https://some.namespace"

    relationship = parse_implicit_relationship(parent_node, relationship_type, relationship_node, graph, doc_namespace)

    assert relationship.spdx_element_id == spdx_element_id
    assert relationship.relationship_type == relationship_type
    assert relationship.related_spdx_element_id == related_spdx_element_id



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

import pytest

from spdx_tools.spdx.model import ExtractedLicensingInfo


def test_correct_initialization():
    extracted_licensing_info = ExtractedLicensingInfo("id", "text", "name", ["reference"], "comment")
    assert extracted_licensing_info.license_id == "id"
    assert extracted_licensing_info.extracted_text == "text"
    assert extracted_licensing_info.license_name == "name"
    assert extracted_licensing_info.cross_references == ["reference"]
    assert extracted_licensing_info.comment == "comment"


def test_wrong_type_in_license_id():
    with pytest.raises(TypeError):
        ExtractedLicensingInfo(license_id=42)


def test_wrong_type_in_extracted_text():
    with pytest.raises(TypeError):
        ExtractedLicensingInfo(extracted_text=42)


def test_wrong_type_in_license_name():
    with pytest.raises(TypeError):
        ExtractedLicensingInfo(license_name=42)


def test_wrong_type_in_cross_references():
    with pytest.raises(TypeError):
        ExtractedLicensingInfo(cross_references=[41, 42])


def test_wrong_type_in_comment():
    with pytest.raises(TypeError):
        ExtractedLicensingInfo(comment=42)



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from dataclasses import field
from enum import Enum, auto

from beartype.typing import List, Optional, Union
from license_expression import LicenseExpression

from spdx_tools.common.typing.dataclass_with_properties import dataclass_with_properties
from spdx_tools.common.typing.type_checks import check_types_and_set_values
from spdx_tools.spdx.model import Checksum, SpdxNoAssertion, SpdxNone


class FileType(Enum):
    SOURCE = auto()
    BINARY = auto()
    ARCHIVE = auto()
    APPLICATION = auto()
    AUDIO = auto()
    IMAGE = auto()
    TEXT = auto()
    VIDEO = auto()
    DOCUMENTATION = auto()
    SPDX = auto()
    OTHER = auto()


@dataclass_with_properties
class File:
    name: str
    spdx_id: str
    checksums: List[Checksum]
    file_types: List[FileType] = field(default_factory=list)
    license_concluded: Optional[Union[LicenseExpression, SpdxNoAssertion, SpdxNone]] = None
    license_info_in_file: List[Union[LicenseExpression, SpdxNoAssertion, SpdxNone]] = field(default_factory=list)
    license_comment: Optional[str] = None
    copyright_text: Optional[Union[str, SpdxNoAssertion, SpdxNone]] = None
    comment: Optional[str] = None
    notice: Optional[str] = None
    contributors: List[str] = field(default_factory=list)
    attribution_texts: List[str] = field(default_factory=list)

    # Deprecated properties that should be replaced during parsing:
    # - file dependencies: replace by a DEPENDENCY_OF relationship (or one of the more precise versions)
    # - artifact of (3 properties): replace by an external package reference and a GENERATED_FROM relationship
    #   between the file and this package

    def __init__(
        self,
        name: str,
        spdx_id: str,
        checksums: List[Checksum],
        file_types: List[FileType] = None,
        license_concluded: Optional[Union[LicenseExpression, SpdxNoAssertion, SpdxNone]] = None,
        license_info_in_file: List[Union[LicenseExpression, SpdxNoAssertion, SpdxNone]] = None,
        license_comment: Optional[str] = None,
        copyright_text: Optional[Union[str, SpdxNoAssertion, SpdxNone]] = None,
        comment: str = None,
        notice: Optional[str] = None,
        contributors: List[str] = None,
        attribution_texts: List[str] = None,
    ):
        file_types = [] if file_types is None else file_types
        license_info_in_file = [] if license_info_in_file is None else license_info_in_file
        contributors = [] if contributors is None else contributors
        attribution_texts = [] if attribution_texts is None else attribution_texts
        check_types_and_set_values(self, locals())



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from unittest.mock import MagicMock, call, mock_open, patch

from spdx_tools.spdx.writer.tagvalue.package_writer import write_package
from tests.spdx.fixtures import package_fixture


def test_package_writer():
    package = package_fixture()

    mock: MagicMock = mock_open()
    with patch(f"{__name__}.open", mock, create=True):
        with open("foo", "w") as file:
            write_package(package, file)

    mock.assert_called_once_with("foo", "w")
    handle = mock()
    handle.write.assert_has_calls(
        [
            call("## Package Information\n"),
            call(f"PackageName: {package.name}\n"),
            call(f"SPDXID: {package.spdx_id}\n"),
            call(f"PackageVersion: {package.version}\n"),
            call(f"PackageFileName: {package.file_name}\n"),
            call(f"PackageSupplier: Person: {package.supplier.name} ({package.supplier.email})\n"),
            call(f"PackageOriginator: Person: {package.originator.name} ({package.originator.email})\n"),
            call(f"PackageDownloadLocation: {package.download_location}\n"),
            call("FilesAnalyzed: True\n"),
            call(f"PackageVerificationCode: {package.verification_code.value} (excludes: ./exclude.py)\n"),
            call("PackageChecksum: SHA1: 71c4025dd9897b364f3ebbb42c484ff43d00791c\n"),
            call(f"PackageHomePage: {package.homepage}\n"),
            call(f"PackageSourceInfo: {package.source_info}\n"),
            call(f"PackageLicenseConcluded: {package.license_concluded}\n"),
            call(f"PackageLicenseInfoFromFiles: {package.license_info_from_files[0]}\n"),
            call(f"PackageLicenseInfoFromFiles: {package.license_info_from_files[1]}\n"),
            call(f"PackageLicenseInfoFromFiles: {package.license_info_from_files[2]}\n"),
            call(f"PackageLicenseDeclared: {package.license_declared}\n"),
            call(f"PackageLicenseComments: {package.license_comment}\n"),
            call(f"PackageCopyrightText: {package.copyright_text}\n"),
            call(f"PackageSummary: {package.summary}\n"),
            call(f"PackageDescription: {package.description}\n"),
            call(f"PackageComment: {package.comment}\n"),
            call(
                f"ExternalRef: PACKAGE-MANAGER {package.external_references[0].reference_type} "
                f"{package.external_references[0].locator}\n"
            ),
            call(f"ExternalRefComment: {package.external_references[0].comment}\n"),
            call(f"PackageAttributionText: {package.attribution_texts[0]}\n"),
            call(f"PrimaryPackagePurpose: {package.primary_package_purpose.name}\n"),
            call("ReleaseDate: 2022-11-01T00:00:00Z\n"),
            call("BuiltDate: 2022-11-02T00:00:00Z\n"),
            call("ValidUntilDate: 2022-11-03T00:00:00Z\n"),
        ]
    )



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import TextIO

from spdx_tools.spdx3.model import CreationInfo
from spdx_tools.spdx3.writer.console.console import write_value
from spdx_tools.spdx.datetime_conversions import datetime_to_iso_string


def write_creation_info(creation_info: CreationInfo, text_output: TextIO, indent: bool = True):
    text_output.write("# Creation Information\n")
    write_value("specVersion", str(creation_info.spec_version), text_output, indent)
    write_value("created", datetime_to_iso_string(creation_info.created), text_output, indent)
    for created_by in creation_info.created_by:
        write_value("created by", created_by, text_output, indent)
    for created_using in creation_info.created_using:
        write_value("created using", created_using, text_output, indent)
    write_value("profile", [profile.name for profile in creation_info.profile], text_output, indent)
    write_value("data license", creation_info.data_license, text_output, indent)
    write_value("comment", creation_info.comment, text_output, indent)



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

from typing import List
from unittest import TestCase

import pytest

from spdx_tools.spdx.validation.snippet_validator import validate_snippet, validate_snippet_within_document
from spdx_tools.spdx.validation.validation_message import SpdxElementType, ValidationContext, ValidationMessage
from tests.spdx.fixtures import document_fixture, snippet_fixture


def test_valid_snippet():
    snippet = snippet_fixture()
    validation_messages: List[ValidationMessage] = validate_snippet_within_document(
        snippet, "SPDX-2.3", document_fixture()
    )

    assert validation_messages == []


@pytest.mark.parametrize(
    "snippet_input, expected_message",
    [
        (
            snippet_fixture(byte_range=(-12, 45)),
            "byte_range values must be greater than or equal to 1, but is: (-12, 45)",
        ),
        (
            snippet_fixture(byte_range=(45, 23)),
            "the first value of byte_range must be less than or equal to the second, but is: (45, 23)",
        ),
        (
            snippet_fixture(line_range=(-12, 45)),
            "line_range values must be greater than or equal to 1, but is: (-12, 45)",
        ),
        (
            snippet_fixture(line_range=(45, 23)),
            "the first value of line_range must be less than or equal to the second, but is: (45, 23)",
        ),
    ],
)
def test_invalid_ranges(snippet_input, expected_message):
    validation_messages: List[ValidationMessage] = validate_snippet_within_document(
        snippet_input, "SPDX-2.3", document_fixture()
    )

    expected = ValidationMessage(
        expected_message,
        ValidationContext(
            spdx_id=snippet_input.spdx_id,
            parent_id=document_fixture().creation_info.spdx_id,
            element_type=SpdxElementType.SNIPPET,
            full_element=snippet_input,
        ),
    )

    assert validation_messages == [expected]


def test_v2_2mandatory_fields():
    snippet = snippet_fixture(license_concluded=None, copyright_text=None)

    assert validate_snippet(snippet, "SPDX-2.3") == []

    validation_messages: List[ValidationMessage] = validate_snippet(snippet, "SPDX-2.2")

    context = ValidationContext(spdx_id=snippet.spdx_id, element_type=SpdxElementType.SNIPPET, full_element=snippet)
    mandatory_fields = ["license_concluded", "copyright_text"]
    expected = [ValidationMessage(f"{field} is mandatory in SPDX-2.2", context) for field in mandatory_fields]

    TestCase().assertCountEqual(validation_messages, expected)



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import pytest

from spdx_tools.spdx.constants import DOCUMENT_SPDX_ID
from spdx_tools.spdx.model import Relationship, RelationshipType, SpdxNoAssertion, SpdxNone
from spdx_tools.spdx.parser.error import SPDXParsingError
from spdx_tools.spdx.parser.tagvalue.parser import Parser
from tests.spdx.parser.tagvalue.test_creation_info_parser import DOCUMENT_STR


@pytest.mark.parametrize(
    "relationship_str, expected_relationship",
    [
        (
            "\n".join(
                [f"Relationship: {DOCUMENT_SPDX_ID} DESCRIBES SPDXRef-File", "RelationshipComment: This is a comment."]
            ),
            Relationship(DOCUMENT_SPDX_ID, RelationshipType.DESCRIBES, "SPDXRef-File", "This is a comment."),
        ),
        (
            f"Relationship: {DOCUMENT_SPDX_ID} PATCH_FOR NOASSERTION",
            Relationship(DOCUMENT_SPDX_ID, RelationshipType.PATCH_FOR, SpdxNoAssertion()),
        ),
        (
            "Relationship: SPDXRef-CarolCompression DEPENDS_ON NONE",
            Relationship("SPDXRef-CarolCompression", RelationshipType.DEPENDS_ON, SpdxNone()),
        ),
        (
            "Relationship: DocumentRef-ExternalDocument:SPDXRef-Test DEPENDS_ON DocumentRef:AnotherRef",
            Relationship(
                "DocumentRef-ExternalDocument:SPDXRef-Test", RelationshipType.DEPENDS_ON, "DocumentRef:AnotherRef"
            ),
        ),
    ],
)
def test_parse_relationship(relationship_str, expected_relationship):
    parser = Parser()
    document = parser.parse("\n".join([DOCUMENT_STR, relationship_str]))
    assert document is not None
    assert len(document.relationships) == 1
    relationship = document.relationships[0]
    assert relationship == expected_relationship


@pytest.mark.parametrize(
    "relationship_str, expected_message",
    [
        (
            "Relationship: spdx_id DESCRIBES",
            [
                "Error while parsing Relationship: [\"Relationship couldn't be split in "
                'spdx_element_id, relationship_type and related_spdx_element. Line: 1"]'
            ],
        ),
        (
            "Relationship: spdx_id IS spdx_id",
            ["Error while parsing Relationship: ['Invalid RelationshipType IS. Line: 1']"],
        ),
    ],
)
def test_parse_invalid_relationship(relationship_str, expected_message):
    parser = Parser()
    with pytest.raises(SPDXParsingError) as err:
        parser.parse(relationship_str)

    assert err.value.get_messages() == expected_message



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import TextIO

from spdx_tools.spdx3.model import SpdxDocument
from spdx_tools.spdx3.writer.console.bundle_writer import write_bundle


def write_spdx_document(spdx_document: SpdxDocument, text_output: TextIO):
    text_output.write("## SPDX Document\n")
    write_bundle(spdx_document, text_output, False)



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

from beartype.typing import List, Optional

from spdx_tools.spdx.model import Document, Package, Relationship, RelationshipType
from spdx_tools.spdx.model.relationship_filters import filter_by_type_and_origin, filter_by_type_and_target
from spdx_tools.spdx.validation.checksum_validator import validate_checksums
from spdx_tools.spdx.validation.external_package_ref_validator import validate_external_package_refs
from spdx_tools.spdx.validation.license_expression_validator import (
    validate_license_expression,
    validate_license_expressions,
)
from spdx_tools.spdx.validation.package_verification_code_validator import validate_verification_code
from spdx_tools.spdx.validation.spdx_id_validators import validate_spdx_id
from spdx_tools.spdx.validation.uri_validators import validate_download_location, validate_url
from spdx_tools.spdx.validation.validation_message import SpdxElementType, ValidationContext, ValidationMessage


def validate_packages(
    packages: List[Package], spdx_version: str, document: Optional[Document] = None
) -> List[ValidationMessage]:
    validation_messages: List[ValidationMessage] = []
    if document:
        for package in packages:
            validation_messages.extend(validate_package_within_document(package, spdx_version, document))
    else:
        for package in packages:
            validation_messages.extend(validate_package(package, spdx_version))

    return validation_messages


def validate_package_within_document(
    package: Package, spdx_version: str, document: Document
) -> List[ValidationMessage]:
    validation_messages: List[ValidationMessage] = []
    context = ValidationContext(
        spdx_id=package.spdx_id,
        parent_id=document.creation_info.spdx_id,
        element_type=SpdxElementType.PACKAGE,
        full_element=package,
    )

    for message in validate_spdx_id(package.spdx_id, document):
        validation_messages.append(ValidationMessage(message, context))

    if not package.files_analyzed:
        package_contains_relationships = filter_by_type_and_origin(
            document.relationships, RelationshipType.CONTAINS, package.spdx_id
        )
        contained_in_package_relationships = filter_by_type_and_target(
            document.relationships, RelationshipType.CONTAINED_BY, package.spdx_id
        )

        combined_relationships: List[Relationship] = (
            package_contains_relationships + contained_in_package_relationships
        )

        if combined_relationships:
            validation_messages.append(
                ValidationMessage(
                    f"package must contain no elements if files_analyzed is False, but found {combined_relationships}",
                    context,
                )
            )

    validation_messages.extend(validate_license_expression(package.license_concluded, document, package.spdx_id))

    license_info_from_files = package.license_info_from_files
    if license_info_from_files:
        if not package.files_analyzed:
            validation_messages.append(
                ValidationMessage(
                    f"license_info_from_files must be None if files_analyzed is False, but is: "
                    f"{license_info_from_files}",
                    context,
                )
            )
        else:
            validation_messages.extend(
                validate_license_expressions(license_info_from_files, document, package.spdx_id)
            )

    validation_messages.extend(validate_license_expression(package.license_declared, document, package.spdx_id))

    validation_messages.extend(validate_package(package, spdx_version, context))

    return validation_messages


def validate_package(
    package: Package, spdx_version: str, context: Optional[ValidationContext] = None
) -> List[ValidationMessage]:
    validation_messages: List[ValidationMessage] = []
    if not context:
        context = ValidationContext(
            spdx_id=package.spdx_id, element_type=SpdxElementType.PACKAGE, full_element=package
        )

    download_location = package.download_location
    if isinstance(download_location, str):
        for message in validate_download_location(download_location):
            validation_messages.append(ValidationMessage("package download_location " + message, context))

    homepage = package.homepage
    if isinstance(homepage, str):
        for message in validate_url(homepage):
            validation_messages.append(ValidationMessage("homepage " + message, context))

    verification_code = package.verification_code
    if verification_code:
        if not package.files_analyzed:
            validation_messages.append(
                ValidationMessage(
                    f"verification_code must be None if files_analyzed is False, but is: {verification_code}", context
                )
            )
        else:
            validation_messages.extend(validate_verification_code(verification_code, package.spdx_id))

    validation_messages.extend(validate_checksums(package.checksums, package.spdx_id, spdx_version))

    validation_messages.extend(
        validate_external_package_refs(package.external_references, package.spdx_id, spdx_version)
    )

    if spdx_version == "SPDX-2.2":
        if package.primary_package_purpose is not None:
            validation_messages.append(
                ValidationMessage("primary_package_purpose is not supported in SPDX-2.2", context)
            )
        if package.built_date is not None:
            validation_messages.append(ValidationMessage("built_date is not supported in SPDX-2.2", context))
        if package.release_date is not None:
            validation_messages.append(ValidationMessage("release_date is not supported in SPDX-2.2", context))
        if package.valid_until_date is not None:
            validation_messages.append(ValidationMessage("valid_until_date is not supported in SPDX-2.2", context))

        if package.license_concluded is None:
            validation_messages.append(ValidationMessage("license_concluded is mandatory in SPDX-2.2", context))
        if package.license_declared is None:
            validation_messages.append(ValidationMessage("license_declared is mandatory in SPDX-2.2", context))
        if package.copyright_text is None:
            validation_messages.append(ValidationMessage("copyright_text is mandatory in SPDX-2.2", context))

    return validation_messages



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from unittest import TestCase

import pytest

from spdx_tools.spdx.model import ChecksumAlgorithm
from spdx_tools.spdx.parser.error import SPDXParsingError
from spdx_tools.spdx.parser.jsonlikedict.checksum_parser import ChecksumParser


def test_parse_checksum():
    checksum_parser = ChecksumParser()
    checksum_dict = {"algorithm": "SHA1", "checksumValue": "d6a770ba38583ed4bb4525bd96e50461655d2759"}

    checksum = checksum_parser.parse_checksum(checksum_dict)

    assert checksum.value == "d6a770ba38583ed4bb4525bd96e50461655d2759"
    assert checksum.algorithm == ChecksumAlgorithm.SHA1


def test_parse_invalid_checksum():
    checksum_parser = ChecksumParser()
    checksum_dict = {"algorithm": "SHA", "checksumValue": "d6a770ba38583ed4bb4525bd96e50461655d2759"}

    with pytest.raises(SPDXParsingError) as err:
        checksum_parser.parse_checksum(checksum_dict)

    TestCase().assertCountEqual(
        err.value.get_messages(), ["Error while parsing Checksum: ['Invalid ChecksumAlgorithm: SHA']"]
    )


def test_parse_incomplete_checksum():
    checksum_parser = ChecksumParser()
    checksum_dict = {"algorithm": "SHA1"}

    with pytest.raises(SPDXParsingError):
        checksum_parser.parse_checksum(checksum_dict)



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from abc import ABC, abstractmethod

from beartype.typing import Any, Dict, Generic, Type, TypeVar

from spdx_tools.spdx.casing_tools import snake_case_to_camel_case
from spdx_tools.spdx.jsonschema.json_property import JsonProperty
from spdx_tools.spdx.model import Document

MISSING_IMPLEMENTATION_MESSAGE = "Must be implemented"

T = TypeVar("T")


class TypedConverter(ABC, Generic[T]):
    """
    Base class for all converters between an instance of the tools-python data model and the corresponding dictionary
    representation, following the json schema. The generic type T is the type according to the data model.
    Each converter has several methods:
    - get_json_type and get_data_model_type: return the data model type and the corresponding JsonProperty subclass.
    These methods are abstract in the base class and need to be implemented in subclasses.
    - json_property_name: converts an enum value of a JsonProperty subclass to the corresponding property name in the
    json schema. The default implementation simply converts from snake case to camel case. Can be overridden in case
    of exceptions like "SPDXID".
    - convert: converts an instance of type T (one of the data model types) to a dictionary representation. In some
    cases, the full document is required (see below). The logic should be generic for all types.
    - requires_full_document: indicates whether the full document is required for conversion. Returns False by
    default, can be overridden as needed for specific types.
    - _get_property_value: Retrieves the value of a specific json property from the data model instance. In some
    cases, the full document is required.
    """

    @abstractmethod
    def _get_property_value(self, instance: T, json_property: JsonProperty, document: Document = None) -> Any:
        raise NotImplementedError(MISSING_IMPLEMENTATION_MESSAGE)

    @abstractmethod
    def get_json_type(self) -> Type[JsonProperty]:
        raise NotImplementedError(MISSING_IMPLEMENTATION_MESSAGE)

    @abstractmethod
    def get_data_model_type(self) -> Type[T]:
        raise NotImplementedError(MISSING_IMPLEMENTATION_MESSAGE)

    def json_property_name(self, json_property: JsonProperty) -> str:
        return snake_case_to_camel_case(json_property.name)

    def requires_full_document(self) -> bool:
        return False

    def convert(self, instance: T, document: Document = None) -> Dict:
        if not isinstance(instance, self.get_data_model_type()):
            raise TypeError(
                f"Converter of type {self.__class__} can only convert objects of type "
                f"{self.get_data_model_type()}. Received {type(instance)} instead."
            )
        if self.requires_full_document() and not document:
            raise ValueError(f"Converter of type {self.__class__} requires the full document")

        result = {}
        for property_name in self.get_json_type():
            property_value = self._get_property_value(instance, property_name, document)
            if property_value is None:
                continue
            result[self.json_property_name(property_name)] = property_value
        return result



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import Optional, Union

from spdx_tools.spdx.model.spdx_no_assertion import SpdxNoAssertion
from spdx_tools.spdx.model.spdx_none import SpdxNone


def handle_no_assertion_or_none(field: Union[SpdxNone, SpdxNoAssertion, str], field_name: str) -> Optional[str]:
    if isinstance(field, SpdxNone):
        print(f"{field_name}: Missing conversion for SpdxNone.")
        return None
    if isinstance(field, SpdxNoAssertion):
        return None
    if isinstance(field, str):
        return field



#  SPDX-FileCopyrightText: 2023 spdx contributors
#
#  SPDX-License-Identifier: Apache-2.0
from os import path

from spdx_tools.spdx.graph_generation import export_graph_from_document
from spdx_tools.spdx.model import Document
from spdx_tools.spdx.parser.parse_anything import parse_file

# This example demonstrates how to generate a relationship graph for an SPDX2 document

# Provide a path to the input file
input_path = path.join(path.dirname(__file__), "..", "tests", "spdx", "data", "SPDXJSONExample-v2.3.spdx.json")
# Parse the file
document: Document = parse_file(input_path)
# Generate the graph (note: you need to have installed the optional dependencies networkx and pygraphviz)
export_graph_from_document(document, "graph.png")



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import List


class SPDXParsingError(Exception):
    messages: List[str]

    def __init__(self, messages: List[str]):
        self.messages = messages

    def get_messages(self):
        return list(self.messages)



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

from beartype.typing import List

from spdx_tools.spdx.model import ExternalDocumentRef
from spdx_tools.spdx.validation.checksum_validator import validate_checksum
from spdx_tools.spdx.validation.spdx_id_validators import is_valid_external_doc_ref_id
from spdx_tools.spdx.validation.uri_validators import validate_uri
from spdx_tools.spdx.validation.validation_message import SpdxElementType, ValidationContext, ValidationMessage


def validate_external_document_refs(
    external_document_refs: List[ExternalDocumentRef], parent_id: str, spdx_version: str
) -> List[ValidationMessage]:
    validation_messages = []
    for external_document_ref in external_document_refs:
        validation_messages.extend(validate_external_document_ref(external_document_ref, parent_id, spdx_version))

    return validation_messages


def validate_external_document_ref(
    external_document_ref: ExternalDocumentRef, parent_id: str, spdx_version: str
) -> List[ValidationMessage]:
    validation_messages = []
    context = ValidationContext(
        parent_id=parent_id, element_type=SpdxElementType.EXTERNAL_DOCUMENT_REF, full_element=external_document_ref
    )

    if not is_valid_external_doc_ref_id(external_document_ref.document_ref_id):
        validation_messages.append(
            ValidationMessage(
                f'document_ref_id must only contain letters, numbers, ".", "-" and "+" and must begin with '
                f'"DocumentRef-", but is: {external_document_ref.document_ref_id}',
                context,
            )
        )

    for message in validate_uri(external_document_ref.document_uri):
        validation_messages.append(ValidationMessage("document_uri " + message, context))

    validation_messages.extend(validate_checksum(external_document_ref.checksum, parent_id, spdx_version))

    return validation_messages



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

import pytest

from spdx_tools.spdx.model import Actor, ActorType


def test_correct_initialization():
    actor = Actor(ActorType.TOOL, "tool_name", "mail")
    assert actor.actor_type == ActorType.TOOL
    assert actor.name == "tool_name"
    assert actor.email == "mail"


def test_correct_initialization_with_optional_as_none():
    actor = Actor(ActorType.TOOL, "tool_name", None)
    assert actor.actor_type == ActorType.TOOL
    assert actor.name == "tool_name"
    assert actor.email is None


def test_wrong_type_in_actor_type():
    with pytest.raises(TypeError):
        Actor("PERSON", "name")


def test_wrong_type_in_name():
    with pytest.raises(TypeError):
        Actor(ActorType.PERSON, 42)


def test_wrong_type_in_email():
    with pytest.raises(TypeError):
        Actor(ActorType.PERSON, "name", [])


def test_wrong_type_in_email_after_initializing():
    with pytest.raises(TypeError):
        actor = Actor(ActorType.PERSON, "name")
        actor.email = []


@pytest.mark.parametrize(
    "actor,expected_string",
    [
        (Actor(ActorType.PERSON, "personName"), "Person: personName"),
        (Actor(ActorType.PERSON, "personName", "personEmail"), "Person: personName (personEmail)"),
        (Actor(ActorType.ORGANIZATION, "orgName"), "Organization: orgName"),
        (Actor(ActorType.ORGANIZATION, "orgName", "orgEmail"), "Organization: orgName (orgEmail)"),
        (Actor(ActorType.TOOL, "toolName"), "Tool: toolName"),
        (Actor(ActorType.TOOL, "toolName", "toolEmail"), "Tool: toolName (toolEmail)"),
    ],
)
def test_serialization(actor: Actor, expected_string: str):
    assert actor.to_serialized_string() == expected_string



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from spdx_tools.spdx3.model.licensing.license_field import LicenseField


class NoneLicense(LicenseField):
    def __init__(self):
        pass



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from dataclasses import fields

from beartype.typing import Any, Dict

from spdx_tools.common.typing.constructor_type_errors import ConstructorTypeErrors


def check_types_and_set_values(instance_under_construction: Any, local_variables: Dict) -> None:
    """
    Helper method to accumulate all type errors encountered during a constructor call and return them in a
    ConstructorTypeErrors instance.
    Background: Our setters are enhanced with runtime typechecks using beartype. However, this means that by
    default, a TypeError is raised on the first type violation that is encountered. We consider it more helpful to
    return all type violations in one go.
    As an aside, defining constructors "manually" using this utility method helps avoid a nasty PyCharm bug:
    https://youtrack.jetbrains.com/issue/PY-34569
    """
    errors = []
    for field in fields(instance_under_construction):
        key = field.name
        value = local_variables.get(key)
        try:
            setattr(instance_under_construction, key, value)
        except TypeError as err:
            error_message: str = err.args[0]
            errors.append(error_message)
    if errors:
        raise ConstructorTypeErrors(errors)



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from dataclasses import field
from enum import Enum, auto

from beartype.typing import List, Optional

from spdx_tools.common.typing.dataclass_with_properties import dataclass_with_properties
from spdx_tools.common.typing.type_checks import check_types_and_set_values


class ExternalIdentifierType(Enum):
    CPE22 = auto()
    CPE23 = auto()
    CVE = auto()
    EMAIL = auto()
    GITOID = auto()
    PURL = auto()
    SECURITY_OTHER = auto()
    SWHID = auto()
    SWID = auto()
    URL_SCHEME = auto()
    OTHER = auto()


@dataclass_with_properties
class ExternalIdentifier:
    external_identifier_type: ExternalIdentifierType
    identifier: str
    comment: Optional[str] = None
    identifier_locator: List[str] = field(default_factory=list)
    issuing_authority: Optional[str] = None

    def __init__(
        self,
        external_identifier_type: ExternalIdentifierType,
        identifier: str,
        comment: Optional[str] = None,
        identifier_locator: List[str] = None,
        issuing_authority: Optional[str] = None,
    ):
        identifier_locator = [] if identifier_locator is None else identifier_locator
        check_types_and_set_values(self, locals())



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
""" This is a temporary package to write the implemented model of spdx_tools.spdx3.0 to console. As soon as
    serialization formats are properly defined this package can be deleted."""



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from datetime import datetime

import pytest
from semantic_version import Version

from spdx_tools.spdx3.bump_from_spdx2.actor import bump_actor
from spdx_tools.spdx3.model import (
    CreationInfo,
    ExternalIdentifier,
    ExternalIdentifierType,
    Organization,
    Person,
    ProfileIdentifier,
    Tool,
)
from spdx_tools.spdx3.payload import Payload
from spdx_tools.spdx.model.actor import Actor, ActorType


@pytest.mark.parametrize(
    "actor_type, actor_name, actor_mail, element_type, new_spdx_id",
    [
        (ActorType.PERSON, "person name", "person@mail.com", Person, "SPDXRef-Actor-personname-person@mail.com"),
        (
            ActorType.ORGANIZATION,
            "organization name",
            "organization@mail.com",
            Organization,
            "SPDXRef-Actor-organizationname-organization@mail.com",
        ),
        (ActorType.TOOL, "tool name", None, Tool, "SPDXRef-Actor-toolname"),
    ],
)
def test_bump_actor(actor_type, actor_name, actor_mail, element_type, new_spdx_id):
    payload = Payload()
    document_namespace = "https://doc.namespace"
    creation_info = CreationInfo(Version("3.0.0"), datetime(2022, 1, 1), ["Creator"], [ProfileIdentifier.CORE])
    actor = Actor(actor_type, actor_name, actor_mail)

    agent_or_tool_id = bump_actor(actor, payload, document_namespace, creation_info)
    agent_or_tool = payload.get_element(agent_or_tool_id)

    assert isinstance(agent_or_tool, element_type)
    assert agent_or_tool.spdx_id == f"{document_namespace}#{new_spdx_id}"
    assert agent_or_tool.name == actor_name
    if actor_mail:
        assert len(agent_or_tool.external_identifier) == 1
        assert agent_or_tool.external_identifier[0] == ExternalIdentifier(ExternalIdentifierType.EMAIL, actor_mail)
    else:
        assert len(agent_or_tool.external_identifier) == 0


def test_bump_actor_that_already_exists():
    creation_info_old = CreationInfo(Version("3.0.0"), datetime(2022, 1, 1), ["Creator"], [ProfileIdentifier.CORE])
    creation_info_new = CreationInfo(Version("3.0.0"), datetime(2023, 2, 2), ["Creator"], [ProfileIdentifier.CORE])

    name = "some name"
    document_namespace = "https://doc.namespace"
    payload = Payload(
        {
            "https://doc.namespace#SPDXRef-Actor-somename-some@mail.com": Person(
                "https://doc.namespace#SPDXRef-Actor-somename-some@mail.com", creation_info_old, name
            )
        }
    )

    actor = Actor(ActorType.PERSON, name, "some@mail.com")
    agent_spdx_id = bump_actor(actor, payload, document_namespace, creation_info_new)

    # assert that there is only one Person in the payload
    assert (
        len(
            [
                payload.get_element(person_id)
                for person_id in payload.get_full_map()
                if isinstance(payload.get_element(person_id), Person)
            ]
        )
        == 1
    )
    assert payload.get_element(agent_spdx_id).creation_info == creation_info_old



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from datetime import datetime


def datetime_from_str(date_str: str) -> datetime:
    if not isinstance(date_str, str):
        raise TypeError(f"Could not convert str to datetime, invalid type: {type(date_str).__name__}")

    date = datetime.strptime(date_str, "%Y-%m-%dT%H:%M:%SZ")  # raises ValueError if format does not match
    return date


def datetime_to_iso_string(date: datetime) -> str:
    """
    Return an ISO-8601 representation of a datetime object.
    """
    if date.microsecond != 0:
        date = date.replace(microsecond=0)  # SPDX does not support microseconds

    return date.isoformat() + "Z"



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from rdflib import RDF, Graph, URIRef

from spdx_tools.spdx.model import ExternalDocumentRef
from spdx_tools.spdx.rdfschema.namespace import SPDX_NAMESPACE
from spdx_tools.spdx.writer.rdf.checksum_writer import add_checksum_to_graph


def add_external_document_ref_to_graph(
    external_document_ref: ExternalDocumentRef, graph: Graph, doc_node: URIRef, doc_namespace: str
):
    external_document_ref_resource = URIRef(f"{doc_namespace}#{external_document_ref.document_ref_id}")
    graph.add((external_document_ref_resource, RDF.type, SPDX_NAMESPACE.ExternalDocumentRef))
    graph.add(
        (external_document_ref_resource, SPDX_NAMESPACE.spdxDocument, URIRef(external_document_ref.document_uri))
    )
    add_checksum_to_graph(external_document_ref.checksum, graph, external_document_ref_resource)

    graph.add((doc_node, SPDX_NAMESPACE.externalDocumentRef, external_document_ref_resource))



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from enum import auto

from spdx_tools.spdx.jsonschema.json_property import JsonProperty


class AnnotationProperty(JsonProperty):
    ANNOTATION_DATE = auto()
    ANNOTATION_TYPE = auto()
    ANNOTATOR = auto()
    COMMENT = auto()



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from spdx_tools.spdx.formats import FileFormat, file_name_to_format
from spdx_tools.spdx.model import Document
from spdx_tools.spdx.writer.json import json_writer
from spdx_tools.spdx.writer.rdf import rdf_writer
from spdx_tools.spdx.writer.tagvalue import tagvalue_writer
from spdx_tools.spdx.writer.xml import xml_writer
from spdx_tools.spdx.writer.yaml import yaml_writer


def write_file(document: Document, file_name: str, validate: bool = True):
    output_format = file_name_to_format(file_name)
    if output_format == FileFormat.JSON:
        json_writer.write_document_to_file(document, file_name, validate)
    elif output_format == FileFormat.YAML:
        yaml_writer.write_document_to_file(document, file_name, validate)
    elif output_format == FileFormat.XML:
        xml_writer.write_document_to_file(document, file_name, validate)
    elif output_format == FileFormat.TAG_VALUE:
        tagvalue_writer.write_document_to_file(document, file_name, validate)
    elif output_format == FileFormat.RDF_XML:
        rdf_writer.write_document_to_file(document, file_name, validate)



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import List

from spdx_tools.common.typing.dataclass_with_properties import dataclass_with_properties
from spdx_tools.common.typing.type_checks import check_types_and_set_values
from spdx_tools.spdx3.model.licensing.any_license_info import AnyLicenseInfo


@dataclass_with_properties
class ConjunctiveLicenseSet(AnyLicenseInfo):
    member: List[AnyLicenseInfo]

    def __init__(self, member: List[AnyLicenseInfo]):
        check_types_and_set_values(self, locals())
